:userdoc.
:title.Documentation for encryption modules
:docprof toc=123.

.***********************************
.*   INTRODUCTION
.***********************************

:h1.Introduction

:p.
This package contains several Modula-2 modules implementing a number
of different encryption algorithms. Some of this was originally part
of the PMOS/2 library, but it turned out to be convenient to move all
encryption functions into a separate package.
:p.
It is distributed as open-source freeware subject to the GNU GPL
licence. Source code is included.

:p.
:hp2.Disclaimer of Warranty:ehp2.

:sl compact.
:li.
:hp1.
This Product is provided "as-is", without warranty of any
kind, either expressed or implied, including, but not limited to,
the implied warranties of merchantability and fitness for a
particular purpose. The entire risk as to the quality and
performance of the Product is with you. Should the Product prove
defective, the full cost of repair, servicing, or correction lies
with you.
:ehp1.
:esl.

:p.
The author of this software is Peter Moylan, peter@pmoylan.org. However, some
of it consists of translations from C of software that was already available.
The individual modules contain copyright information where relevant.

.***********************************
.*   WHAT IS INCLUDED
.***********************************

:h1 id=overview.Package overview

:hp2.Package overview:ehp2.

:p.What you have here is a collection of Modula-2 source code for a
variety of encryption algorithms. If you don't want to use Modula-2, it
should be easy enough to translate the code into other languages. I prefer
to use Modula-2 for ease of program development and efficiency of object
code, but also because the source code is very readable. (I've tried
looking at C source code for some of these algorithms, and most of it has
been so difficult to read that I gave up; it's tempting to say that the sources
were encrypted.) That readability is why it should be easy to translate
the code into other languages.

:p.There are, by the way, some Modula-to-C translators available. They
are pretty good in terms of producing useful C code, but they tend to do
things, like deleting comments, that hurt the readability.

:p.Modula-2 programs are broken up into modules in such a way that you
shouldn't need to understand the details of other modules in order to
understand the current module. That is because the information-hiding
walls between modules are fairly strong. (Unlike C++, for example, where
typically the internal fields of an object are revealed in a header file.)
Every module, except the top-level one, is made up of two parts: the
implementation module and the definition module. The implementation module
contains the details. The definition module is the interface; the part
that a programmer needs in order to use that module. The definition module
contains only procedure declarations (but not the procedure body) and a
few things like that. Most importantly, it contains the comments that
say what each procedure does. Someone who wants to call that procedure
shouldn't have to look inside the implementation module. That is detail
that should be irrelevant to the caller.

:p.In this package, the DEF directory contains the definition modules,
and the SRC directory contains the implementation modules. (So you
look into the DEF directory to see what is included, and you look into
the SRC directory if you want to see how it was implemented.) The SRC
directory contains a subdirectory called Tests that contain the top-level
programs that test these modules.

:p.For completeness, I have included a file xc.red, which is what my
compiler uses to define its search paths. That says where to look for
the various file types.

.***********************************
.*   OVERVIEW OF ENCRYPTION
.***********************************

:h1 id=encryptionoverview.Overview of encryption methods

:p.:hp2.Overview of encryption methods:ehp2.

:p.Encryption algorithms can be broadly divided into two classes:
one-way encryption and two-way encryption. These can be further
subdivided into a few subclasses.

:p.:hp2.One-way encryption without a key:ehp2.

:p.The simplest case is where we encrypt the data with an irreversible
algorithm. (One that is known to both sender and receiver.)
There is no way to decrypt the result. Typically this
takes input data of arbitrary length and produces a short fixed-length
"digest" as the result. This is used for signing a document, with evidence
that it has not been altered. Both the original document and the digest
are sent to the receiver. The receiver recomputes the digest, and compares
it with the original digest. If they are different, the document has been
altered.

:p.The non-keyed one-way encryption methods included in this package are
MD5, SHA1, SHA224, SHA256, SHA384, SHA512.

:p.:hp2.One-way encryption with a key:ehp2.

:p.This is similar to the above case, except that the encryption result
also depends on a secret key, the key usually being a short byte string.
The key is known to both the sender and the receiver, but possibly not
to third parties. The receiver repeats the encryption calculation, and
should get the same result. One application of this is for sending a
password in encrypted form. The password is known to both sender and
receiver, but not (we hope) to anyone else.

:p.This can work even if the key can be intercepted by a third party,
provided that the key is changed often enough. (It might, for example,
be a timestamp.) An attacker can intercept an encrypted password, but
cannot reuse it unless the key has not yet been changed.

:p.The keyed one-way encryption methods included in this package are
HMAC_MD5, HMAC_SHA1, HMAC_SHA256, HMAC_SHA384, HMAC_SHA512.

:p.:hp2.Block and stream ciphers:ehp2.

:p.These are reversible encodings, which depend on a secret key that is
known only to the sender and receiver. Where one-way encryption methods are
restricted to special cases like sending passwords or digital signing,
reversible ciphers can be used to transmit texts of arbitrary length, with
the property that they can be deciphered at the receiving end.

:p.The difference between the two is that a stream cipher works a byte at
a time, or in some cases even a bit at a time, continuously transforming the input.
The current output depends on the current input and a current internal state.
The state itself depends on past inputs. Block ciphers
group the input into fixed-length (usually short) blocks, and produce output of
the same block size. The last block is padded at the end, if necessary, with some
dummy data so as to produce an integral number of blocks.

:p.One weakness of block ciphers is that if two or more blocks have the same
input then they will have the same output, and an attacker might be able
to get information from the repetition. One way around this is to make the
ciphertext output depend on the past state as well as the current input.
Two popular ways to do this are to use the so-called CBC or CTR modes of a block cipher.

:p.A feature of all these methods is that the :hp1.same:ehp1. secret key is usually used
for both encryption and decryption. Of course, we must arrange that the key is
exchanged in a way that cannot be intercepted by a third party.

:p.The cipher algorithms included in this package are RC4, DES, DES3,
AES128, AES192, AES256, including the DES options called DES-CBC and DES3-CBC
and the AES options called AES-CBC and AES-CTR.
These should be sufficient for transport layer security.  Of these, RC4 is
considered too vulnerable to attacks to be worth using, but is included here
because some software still uses it.
The DES algorithm is included, but not recommended because of doubts about its security.
The DES variant called 3DES or Triple-DES is stronger, and is also included.

:p.:hp2.Public key encryption:ehp2.

:p.In public key encryption, two keys are used, and they are known as the
public key and the private key. The public key is available for anyone to discover,
but the private key is kept secret. Anyone wanting to send data to the receiver
will encrypt it with the receiver's public key. The receiver then decrypts it
using its private key.

:p.This appears to make all of the above methods obsolete, but there is a catch.
Public key encryption systems are mathematically complicated - they have to be, to
keep the encrypted data secret from an attacker - and that makes them use a lot of
processor time. In practice, we use public key encryption only for as long as is
needed to transmit the data, including the secret key, that is needed to use a
simpler block or stream cipher. Then that cipher is used for the rest of the
communication.

:p.The only public key encryption method included so far is RSA.

.***********************************
.*   WHAT IS INCLUDED
.***********************************

:h1 id=whatitcontains.What has been implemented?

:hp2.What has been implemented:ehp2.

:p.Implementations of the following algorithms are included.

:ul.
:li.MD5.
:li.SHA1, SHA224, SHA256, SHA384, SHA512.
:li.HMAC_MD5, HMAC_SHA1, HMAC_SHA256, HMAC_SHA384, HMAC_SHA512.
:li.DES_ECB, DES3_ECB, DES_CBC, DES3_CBC.
:li.AES128, AES192, AES256, and also AES_CBC128, AES_CBC192, AES_CBC256,
AES_CTR128, AES_CTR192, AES_CTR256.
:li.RSA.
:eul.

:p.:hp2.MD5:ehp2.

:p.MD5 stands for message digest, version 5. (Nobody uses the earlier
versions.) This takes an arbitrarily long input string - sometimes an
entire file - and produces an 128-bit digest. This is, of course, a one-way
encryption - that is, the original text cannot be recovered from the digest -
and is used mainly as a sort of signature for a file. The sender and
receiver both compute the digest, and if they don't agree then the file
has been altered in transmission.

:p.A "keyed" version known as CRAM-MD5 is still in common use for mail
password authentication. (CRAM-MD5 is not included in the current package,
but may be found in the sources for my Weasel mail server.) Here the key,
which is a timestamp provided by the server, is combined with the password,
and the combination is MD5 encrypted. Both the server and the client do
the encryption, and if the password is valid then the results should agree.

:p.:hp2.SHA1, SHA224, SHA256, SHA384, SHA512:ehp2.

:p.SHA stands for "secure hash algorithm". It, like MD5, is a one-way
encryption, and it is used for similar purposes, but it is intended to be
more secure against attacks than MD5. The different numbers attached to the
SHA algorithms refer to increasing levels of security, with increasing version
numbers giving longer digests.

:p.In this package, SHA1 is contained in the module called SHA1. SHA224 and SHA256
are in the module called SHA2, and SHA384 and SHA512 are in the SHA512 module.

:p.The various SHA functions, along with MD5, are considered to be
"cryptographic hash functions". That is because they map a string of arbitrary
length into a fixed-length pseudo-random number.

:p.They are also stream ciphers, in the sense that the current output depends
on past inputs. They do process the data in fixed-length blocks, but the
result for the current block depends on a "state" that depends on past blocks.

:p.:hp2.HMAC:ehp2.

:p.HMAC is a "keyed" version of MD5 and the various versions of SHA. The CRAM-MD5
algorithm, mentioned earlier, also uses a key, but in that case the key and the
message text are simply concatenated before doing the encryption. HMAC uses a
more complicated way of combining the key and the message text, which should
improve security. In principle the HMAC algorithm can be used with any
cryptographic hash function, but in practice I have found it implemented only
with MD5 or the SHA family.

:p.All of the variants I know about are included in the module called HMAC.

:p.:hp2.RC4:ehp2.

:p.RC4 was originally a trade secret, so what we have here should more strictly
be known as "alleged RC4", also called ARC4, based on leaked information. It
was popular for some time because of its simplicity and speed. It is now
considered to be not sufficiently safe against attacks, but it is included here
for completeness.

:p.RC4 is deprecated for use in TLS (transport layer security) by RFC 7465.

:p.:hp2.DES and 3DES:ehp2.

:p.DES, standing for Data Encryption Standard, used to be the preferred
reversible encryption method, but it is falling out of fashion because of
suspicions that it is not cryptographically strong enough. A stronger variant called
DES3 or 3DES or triple-DES is still in use. In the present implementation, the
decision as to whether to use DES or DES3 is made internally, based on the
length of the supplied key. The CBC versions of DES and DES3 are also included.

:p.:hp2.AES:ehp2.

:p.AES, also known as the Rijndael cipher after its inventors, stands for "Advanced encryption standard".
Like DES, but unlike the other methods listed above, this is an invertible cipher. It consists
of a sequence of steps, each of which is invertible. Some of the steps depend on
a secret key, so the inversion is possible only if one knows the key. That means that
it can be used for secure transmission between two points in a network, provided
that the key is known at both ends.

:p.A number of invertible ciphers have been proposed, but of these the AES method
so far appears to be the most invulnerable against attack.

:p.AES is a block cipher, in the following sense: the translation works on 16-byte
blocks at a time, and the result does not depend on the state of previous blocks.
That would appear to be a security risk, but so far AES has proved to be secure
against attack. To remove even that problem, a CBC mode for AES is also included.

:p.There are three variants of AES, called AES128, AES192, and AES256, with increasingly
longer keys. All of these are implemented in the module called AES.

:p.AES is symmetric in the sense that the same key is used for encryption and
decryption. That feature can also be found in the Data Encryption Standard (DES). There are
also several known public-key cryptography methods, using two keys where one is
public and the other is private, but none of those are included in the current
package.

:p.:hp2.RSA:ehp2.

:p.RSA is the best-known public key algorithm. An associated module RSAGenerate can
be used to create RSA keys.

.***********************************
.*   STILL TO BE IMPLEMENTED
.***********************************

:h1 id=notyetimplemented.Not yet implemented

:hp2.Not yet implemented:ehp2.

:p.The following algorithms are under consideration for being added to this package.

:ul.
:li. P_sha256, see RFC 5246, section 5, pp 14-15.
:eul.

:euserdoc.

