    (**************************************************************************)
(*                                                                        *)
(*  Decoding an ASN.1 coded file                                          *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE ASN1decode;

        (********************************************************)
        (*                                                      *)
        (*              Decoding an ASN.1 coded file            *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 September 2018               *)
        (*  Last edited:        2 October 2018                  *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, TextIO, IOChan, BigNum;

FROM SYSTEM IMPORT CARD8;

FROM ASN1 IMPORT
    (* proc *)  ASN1toTxt;

FROM VarStrings IMPORT
    (* type *)  VarStringPtr,
    (* proc *)  WriteVarString;

FROM BigNum IMPORT BN;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, OpenNewFile, CloseFile, Exists, DeleteFile,
                ReadLine, ReadRaw, WriteRaw;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Base64 IMPORT
    (* proc *)  Decode;

FROM LowLevel IMPORT
    (* proc *)  IAND, IANDB;

(************************************************************************)

TYPE FilenameString = ARRAY [0..511] OF CHAR;

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = "NPC.PEM";

(************************************************************************)
(*                     MISCELLANOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE StringMatch (str1, str2: ARRAY OF CHAR): BOOLEAN;

    (* Checks if str1 and str2 are equal, modulo character case. *)

    BEGIN
        Strings.Capitalize (str1);
        Strings.Capitalize (str2);
        RETURN Strings.Equal (str1, str2);
    END StringMatch;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);
    END GetParameter;

(************************************************************************)
(*                          LOAD BINARY FILE                            *)
(************************************************************************)

PROCEDURE LoadFile (name: ARRAY OF CHAR;  VAR (*OUT*) buffer: ARRAY OF CARD8): CARDINAL;

    VAR cid: ChanId;
        N: CARDINAL;

    BEGIN
        cid := OpenOldFile (name, FALSE, TRUE);
        ReadRaw (cid, buffer, HIGH(buffer), N);
        CloseFile (cid);
        RETURN N;
    END LoadFile;

(************************************************************************)
(*                        TRANSLATE PEM TO BINARY                       *)
(************************************************************************)

PROCEDURE ReadPEM (VAR (*IN*) filename: FilenameString;
                                VAR (*OUT*) buffer: ARRAY OF CHAR);

    (* Loads a Base64 string from a file, deleting any line that    *)
    (* begins with the three characters '---'.                      *)

    VAR line: ARRAY [0..511] OF CHAR;
        cid: ChanId;

    BEGIN
        buffer[0] := Nul;
        cid := OpenOldFile (filename, FALSE, FALSE);
        LOOP
            ReadLine (cid, line);
            IF line[0] = CtrlZ THEN
                EXIT (*LOOP*);
            ELSIF (line[0] <> '-') AND (line[1] <> '-') AND (line[2] <> '-') THEN
                Strings.Append(line, buffer);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
    END ReadPEM;

(************************************************************************)

PROCEDURE PEMtoBinary (VAR (*IN*) infilename: FilenameString;
                        VAR (*OUT*) bindata: ARRAY OF CARD8): CARDINAL;

    (* Reads a PEM file, translates the data to binary in bindata.      *)
    (* Also writes bindata to an output file.  Returns the number of    *)
    (* output bytes.                                                    *)

    CONST
        InBufSize = 65536;

    VAR outfilename: FilenameString;
        inbuf: ARRAY [0..InBufSize-1] OF CHAR;
        N, pos: CARDINAL;  cid: ChanId;
        found: BOOLEAN;

    BEGIN
        (* Create the output file name from the input name. *)

        Strings.Assign (infilename, outfilename);
        Strings.FindPrev ('.', outfilename, LENGTH(outfilename) - 1,
                                                            found, pos);
        IF found THEN
            outfilename[pos] := Nul;
        END (*IF*);
        Strings.Append (".bin", outfilename);
        DeleteFile (outfilename);

        (* Do the translation. *)

        ReadPEM (infilename, inbuf);
        N := Decode (inbuf, bindata);
        IF N <> 0 THEN
            WriteString ("Creating file ");
            WriteString (outfilename);  WriteLn;
            cid := OpenNewFile (outfilename, TRUE);
            WriteRaw (cid, bindata, N);
            CloseFile (cid);
        END (*IF*);

        RETURN N;

    END PEMtoBinary;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoDecode;

    CONST BufSize = 65536;

    VAR infilename: FilenameString;
        binbuff: ARRAY [0..BufSize-1] OF CARD8;
        extension: ARRAY [0..15] OF CHAR;
        txtptr: VarStringPtr;
        M, N, pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        (* Get the input file name. *)

        GetParameter (infilename);
        N := Strings.Length (infilename);
        WHILE (N > 0) AND (infilename[N-1] = ' ') DO
            DEC(N);
        END (*WHILE*);
        infilename[N] := Nul;
        IF infilename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        END (*IF*);

        (* What sort of file is it?  Does it exist?  *)

        IF NOT Exists (infilename) THEN
            Strings.Append (".bin", infilename);
        END (*IF*);
        Strings.FindPrev ('.', infilename, LENGTH(infilename) - 1,
                                                            found, pos);
        IF found THEN
            Strings.Extract (infilename, pos+1, SIZE(extension), extension);
        ELSE
            extension := "";
        END (*IF*);
        IF NOT Exists (infilename) THEN
            infilename[pos] := Nul;
            Strings.Append (".pem", infilename);
            extension := "PEM";
        END (*IF*);

        (* Do the decoding.  *)

        WriteString ("Reading from file ");
        WriteString (infilename);  WriteLn;
        IF StringMatch (extension, "PEM") THEN
            N := PEMtoBinary (infilename, binbuff);
        ELSE
            N := LoadFile (infilename, binbuff);
        END (*IF*);
        IF N = 0 THEN
            WriteString ("Could not load input file ");
            WriteString (infilename);
            WriteLn;
        ELSE
            WriteCard (N);  WriteString (" bytes total");  WriteLn;
            pos := 0;
            M := ASN1toTxt (binbuff, N, txtptr);
            WriteVarString (txtptr^, M);
        END (*IF*);

    END DoDecode;

(************************************************************************)

BEGIN
    DoDecode;
END ASN1decode.

