(**************************************************************************)
(*                                                                        *)
(*  Decoding an ASN.1 coded file                                          *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE ASN1decode;

        (********************************************************)
        (*                                                      *)
        (*              Decoding an ASN.1 coded file            *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            27 May 2018                     *)
        (*  Last edited:        6 September 2018                *)
        (*  Status:             Partly written                  *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, TextIO, IOChan, BigNum;

FROM SYSTEM IMPORT CARD8;

FROM BigNum IMPORT BN;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, OpenNewFile, CloseFile, Exists, DeleteFile,
                ReadLine, ReadRaw, WriteRaw;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Base64 IMPORT
    (* proc *)  Decode;

FROM LowLevel IMPORT
    (* proc *)  IAND, IANDB;

(************************************************************************)

TYPE FilenameString = ARRAY [0..511] OF CHAR;

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = "NPC2.pem";

(************************************************************************)
(*                     MISCELLANOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE StringMatch (str1, str2: ARRAY OF CHAR): BOOLEAN;

    (* Checks if str1 and str2 are equal, modulo character case. *)

    BEGIN
        Strings.Capitalize (str1);
        Strings.Capitalize (str2);
        RETURN Strings.Equal (str1, str2);
    END StringMatch;

(************************************************************************)

PROCEDURE Spaces (amount: CARDINAL);

    (* Writes amount space characters. *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 1 TO amount DO
            WriteChar (' ');
        END (*FOR*);
    END Spaces;

(********************************************************************************)

PROCEDURE WriteHexChar (N: CARD8);

    (* Write a single hexadecimal digit.  *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(ORD('A') + (N - 10)));
        ELSE
            WriteChar (CHR(ORD('0') + N));
        END (*IF*);
    END WriteHexChar;

(************************************************************************)

PROCEDURE WriteHex (N: CARD8);

    (* Write N in hexadecimal.  *)

    BEGIN
        WriteHexChar (N DIV 16);
        WriteHexChar (N MOD 16);
    END WriteHex;

(************************************************************************)

PROCEDURE WriteHexString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL;
                            initialindent, indent: CARDINAL);

    (* Write N bytes starting at buf[pos].  *)

    VAR j, canfit: CARDINAL;

    BEGIN
        canfit := (80 - initialindent) DIV 3;
        IF N <= canfit THEN
            FOR j := 0 TO N-1 DO
                WriteHex (buf[pos]);  INC(pos);
                WriteChar (' ');
            END (*FOR*);
        ELSE
            WriteLn;
            canfit := (80 - indent) DIV 3;
            WHILE N > 0 DO
                Spaces (indent);
                FOR j := 0 TO canfit-1 DO
                    WriteHex (buf[pos]);  INC(pos);
                    WriteChar (' ');
                END (*FOR*);
                IF N >= canfit THEN
                    DEC (N, canfit);
                ELSE
                    N := 0;
                END (*IF*);
                WriteLn;
            END (*WHILE*);
        END (*IF*);
    END WriteHexString;

(************************************************************************)

PROCEDURE WriteAscString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Like WriteHexString, but we interpret the bytes as characters.  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteChar (CHR(buf[pos]));  INC(pos);
        END (*FOR*);
    END WriteAscString;

(************************************************************************)

PROCEDURE WriteCharString (VAR (*IN*) buf: ARRAY OF CARD8;
                                        pos, N, option: CARDINAL);

    (* Write character string, N bytes starting from buffer[pos].       *)
    (*    option = 1        ASCII                                       *)
    (*    option = 2        UTF8                                        *)

    BEGIN
        (* For now I'm ignoring the option. *)

        IF N > 0 THEN
            WriteAscString (buf, pos, N);
        END (*IF*);

    END WriteCharString;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);
    END GetParameter;

(************************************************************************)
(*                          LOAD BINARY FILE                            *)
(************************************************************************)

PROCEDURE LoadFile (name: ARRAY OF CHAR;  VAR (*OUT*) buffer: ARRAY OF CARD8): CARDINAL;

    VAR cid: ChanId;
        N: CARDINAL;

    BEGIN
        cid := OpenOldFile (name, FALSE, TRUE);
        ReadRaw (cid, buffer, HIGH(buffer), N);
        CloseFile (cid);
        RETURN N;
    END LoadFile;

(************************************************************************)
(*                TRANSLATE BINARY ASN.1 TO HUMAN-READABLE              *)
(************************************************************************)

PROCEDURE ParseASN1 (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL); FORWARD;

    (* Translates ASN.1 code starting at buffer[pos], total N bytes.    *)
    (* We update pos.                                                   *)

(************************************************************************)

PROCEDURE NewLine (indent: CARDINAL);

    (* Starts a new indented line. *)

    BEGIN
        WriteLn;
        Spaces (indent);
    END NewLine;

(************************************************************************)

PROCEDURE WriteInt (val: BN;  length, indent: CARDINAL);

    (* Writes an integer value. *)

    BEGIN
        IF length <= 16 THEN
            BigNum.WriteBignum(val);
        ELSE
            BigNum.WriteBignum(val);
        END (*IF*);
    END WriteInt;

(************************************************************************)

PROCEDURE ParseSEQ (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates a sequence.    *)

    VAR nextpos: CARDINAL;

    BEGIN
        nextpos := pos + N;
        Spaces (indent);  WriteChar ('{');  WriteLn;
        WHILE pos < nextpos DO
            ParseASN1 (buffer, pos, N, indent);
        END (*WHILE*);
        Spaces (indent);  WriteChar ('}');  WriteLn;
    END ParseSEQ;

(************************************************************************)

PROCEDURE ParseSET (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates a set.    *)

    VAR nextpos: CARDINAL;

    BEGIN
        nextpos := pos + N;
        Spaces (indent);  WriteChar ('{');  WriteLn;
        WHILE pos < nextpos DO
            ParseASN1 (buffer, pos, N, indent);
        END (*WHILE*);
        Spaces (indent);  WriteChar ('}');  WriteLn;
    END ParseSET;

(************************************************************************)

PROCEDURE ParseASN1 (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates ASN.1 code starting at buffer[pos], total N bytes.    *)
    (* We update pos.                                                   *)

    (********************************************************************)

    PROCEDURE Fetch(): CARD8;

        (* Gets next byte from buffer. *)

        VAR val: CARD8;

        BEGIN
            val := buffer[pos];  INC(pos);  DEC(N);
            RETURN val;
        END Fetch;

    (********************************************************************)

    PROCEDURE UpdateNum (VAR (*INOUT*) A: BN;  base, nextdigit: CARDINAL);

        (* A := base*A + nextdigit *)

        VAR B, C, D, E: BN;

        BEGIN
            B := BigNum.MakeBignum(base);
            C := BigNum.Prod (A, B);
            D := BigNum.MakeBignum (nextdigit);
            E := BigNum.Sum (C, D);
            BigNum.Discard (A);
            BigNum.Discard (B);
            BigNum.Discard (C);
            BigNum.Discard (D);
            A := E;
        END UpdateNum;

    (********************************************************************)

    PROCEDURE AppendNextByte (VAR (*INOUT*) A: BN);

        (* A := 256*A + Fetch() *)

        BEGIN
            UpdateNum (A, 256, Fetch());
        END AppendNextByte;

    (********************************************************************)

    PROCEDURE GetInt (length: BN): BN;

        (* Picks up an integer of length bytes.  Remark: I should allow *)
        (* for negative numbers in twos complement notation, but so far *)
        (* I haven't done this.                                         *)

        VAR result: BN;

        BEGIN
            result := BigNum.Zero();
            WHILE BigNum.Sign(length) > 0 DO
                AppendNextByte (result);
                BigNum.Decr(length);
            END (*WHILE*);
            RETURN result;
        END GetInt;

    (********************************************************************)

    PROCEDURE Skip (length: CARDINAL;  variable: BOOLEAN);

        (* Jumps over a value we don't want to interpret.  This should  *)
        (* be relevant only for types we don't yet recognise.           *)

        VAR val: CARD8;

        BEGIN
            IF variable THEN
                LOOP
                    val := buffer[pos];  INC (pos);  DEC (N);
                    IF val = 0 THEN
                        val := buffer[pos];  INC (pos);  DEC (N);
                        IF val = 0 THEN EXIT (*LOOP*) END(*IF*);
                    END (*IF*);
                END (*LOOP*);
            ELSE
                INC (pos, length);  DEC (N, length);
            END (*IF*);
        END Skip;

    (********************************************************************)

    VAR type, class, shortlength, j, indentp3: CARDINAL;
        val, length: BN;
        temp: CARD8;
        constructed, last, varlength, ShowDetails: BOOLEAN;

    BEGIN
        indentp3 := indent + 3;

        (* First byte has type information.  In ASN.1 notation this is  *)
        (* called a tag.  Universal tags have numbers in the range 1 to *)
        (* 30 and class=0.  Other values are application-dependent.     *)

        type := Fetch();
        class := type DIV 64;
        constructed := IAND (type, 32) <> 0;
        type := IAND (type, 31);
        IF type = 31 THEN
            type := 0;
            REPEAT
                temp := Fetch();
                last := IAND(temp, 128) = 0;
                type := 128*type + IAND(temp,127);
            UNTIL last;
        END (*IF*);

        (* Second byte has length, or code to indicate long length.     *)
        (* The special case where varlength = TRUE means indefinite     *)
        (* length, in which case the value is terminated by two         *)
        (* consecutive zero bytes.                                      *)

        shortlength := Fetch();
        varlength := shortlength = 128;
        length := BigNum.Zero();
        IF shortlength < 128 THEN
            length := BigNum.MakeBignum (shortlength);
        ELSIF NOT varlength THEN
            (* Long form for length. *)
            temp := shortlength - 128;
            FOR j := 0 TO temp-1 DO
                AppendNextByte (length);
            END (*FOR*);
        END (*IF*);

        Spaces (indent);
        IF class = 0 THEN
            CASE type OF
                    1..6, 12, 16..17, 19, 22..23:
                            ShowDetails := FALSE;
                    ELSE    ShowDetails := TRUE;
            END (*IF*);
        ELSE
            ShowDetails := FALSE;
        END (*IF*);

        IF ShowDetails THEN
            (* Say what we have so far. *)

            WriteString ("class=");  WriteCard(class);
            WriteString (", type=");  WriteCard(type);
            IF constructed THEN WriteString (" constructed") END(*IF*);
            WriteString (", length=");
            IF varlength THEN
                WriteString ("indefinite");
            ELSE
                BigNum.WriteBignum (length);
            END (*IF*);
            WriteChar (' ');
        END (*IF*);

        shortlength := BigNum.Card(length);
        IF class = 1 THEN
            WriteString ("AppType ");  WriteCard (type);
            WriteString (":  ");
            WriteHexString (buffer, pos, shortlength, indent+12, indentp3);
            INC (pos, shortlength);  DEC (N, shortlength);  WriteLn;
        ELSIF class = 2 THEN
            WriteChar ('[');  WriteCard(type);
            WriteString ("] {");  WriteLn;
            ParseASN1 (buffer, pos, shortlength, indentp3);
            Spaces (indentp3);  WriteChar ('}');  WriteLn;
        ELSIF class = 3 THEN
            WriteString ("PrivateType ");  WriteCard (type);
            WriteString (":  ");
            WriteHexString (buffer, pos, shortlength, indent+16, indentp3);
            INC (pos, shortlength);  DEC (N, shortlength);  WriteLn;
        ELSE            (* class 0 is Universal, the most common class *)
            CASE type OF
                |   1:  (* BOOLEAN *)
                        WriteString ("BOOLEAN ");
                        IF Fetch() = 0 THEN
                            WriteString ("FALSE");
                        ELSE
                            WriteString ("TRUE");
                        END (*IF*);
                        WriteLn;

                |   2:  (* INTEGER *)
                        WriteString ("INTEGER ");
                        val := GetInt(length);  WriteInt (val, shortlength, indent);  WriteLn;
                        BigNum.Discard (val);

                |   3:  (* BIT STRING *)
                        (* N.B. constructed encoding has to be handled              *)
                        (* differently, but this should not occur in DER encoding.  *)
                        WriteString ("BIT STRING "); WriteCard (Fetch());
                        DEC (shortlength);  WriteString (" unused bits");
                        WriteHexString (buffer, pos, shortlength, indent+12, indentp3);
                        Skip (shortlength, FALSE);  WriteLn;

                |   4:  (* OCTET STRING *)
                        (* N.B. constructed encoding has to be handled              *)
                        (* differently, but this should not occur in DER encoding.  *)
                        WriteString ("OCTET STRING ");
                        WriteHexString (buffer, pos, shortlength, indent+13, indentp3);
                        Skip (shortlength, FALSE);  WriteLn;

                |   5:  (* NULL *)
                        WriteString ("NULL");  WriteLn;
                        (* Zero content, zero length. *)

                |   6:  (* OBJECT IDENTIFIER *)
                        WriteString ("OBJECT IDENTIFIER {");
                        temp := Fetch();  WriteCard(temp DIV 40);  WriteChar (' ');
                        WriteCard (temp MOD 40);
                        DEC (shortlength);
                        WHILE shortlength > 0 DO
                            val := BigNum.Zero();
                            LOOP
                                j := Fetch();
                                DEC(shortlength);
                                IF IAND (j, 80H) <> 0 THEN
                                    UpdateNum (val, 128, IAND (j, 7FH));
                                ELSE
                                    UpdateNum (val, 128, IAND (j, 7FH));
                                    EXIT (*LOOP*);
                                END (*IF*);
                            END (*LOP*);
                            WriteChar (' ');  BigNum.WriteBignum(val);
                        END (*IF*);
                        WriteChar ('}');  WriteLn;

                |   7:  (* ObjectDescriptor, not yet handled *)
                        WriteString ("ObjectDescriptor, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |   8:  (* INSTANCE OF, EXTERNAL, not yet handled *)
                        WriteString ("INSTANCE OF, EXTERNAL, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |   9:  (* REAL, not yet handled *)
                        WriteString ("REAL, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  10:  (* ENUMERATED, not yet handled *)
                        WriteString ("ENUMERATED, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  11:  (* EMBEDDED PDV, not yet handled *)
                        WriteString ("EMBEDDED PDV, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  12:  (* UTF8String *)
                        WriteString ('UTF8String "');
                        WriteCharString (buffer, pos, shortlength, 2);
                        WriteChar ('"');  WriteLn;
                        Skip (shortlength, FALSE);

                |  13:  (* RELATIVE-OID, not yet handled *)
                        WriteString ("RELATIVE-OID, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  16:  WriteString ("SEQUENCE");  WriteLn;
                        ParseSEQ (buffer, pos, shortlength, indentp3);

                |  17:  WriteString ("SET");  WriteLn;
                        ParseSET (buffer, pos, shortlength, indentp3);

                |  18:  (* NumericString, not yet handled *)
                        WriteString ("NumericString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  19:  (* PrintableString *)
                        (* N.B. constructed encoding has to be handled              *)
                        (* differently, but this should not occur in DER encoding.  *)
                        WriteString ('PrintableString "');
                        WriteCharString (buffer, pos, shortlength, 1);
                        WriteChar ('"');  WriteLn;
                        Skip (shortlength, FALSE);

                |  20:  (* TeletexString, not yet handled *)
                        WriteString ("TeletexString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  21:  (* VideotexString, not yet handled *)
                        WriteString ("VideotexString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  22:  (* IA5String *)
                        (* N.B. constructed encoding has to be handled              *)
                        (* differently, but this should not occur in DER encoding.  *)
                        WriteString ('IA5String "');
                        WriteCharString (buffer, pos, shortlength, 1);
                        WriteChar ('"');  WriteLn;
                        Skip (shortlength, FALSE);

                |  23:  (* UTCTime *)
                        (* N.B. constructed encoding has to be handled              *)
                        (* differently, but this should not occur in DER encoding.  *)
                        WriteString ('UTCTime ');
                        WriteCharString (buffer, pos, shortlength, 1);
                        WriteLn;
                        Skip (shortlength, FALSE);

                |  24:  (* GeneralizedTime, not yet handled *)
                        WriteString ("GeneralizedTime, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  25:  (* GraphicString, not yet handled *)
                        WriteString ("GraphicString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  26:  (* VisibleString, not yet handled *)
                        WriteString ("VisibleString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  27:  (* GeneralString, not yet handled *)
                        WriteString ("GeneralString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  28:  (* UniversalString, not yet handled *)
                        WriteString ("UniversalString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  29:  (* CHARACTER STRING, not yet handled *)
                        WriteString ("CHARACTER STRING, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);

                |  30:  (* BMPString, not yet handled *)
                        WriteString ("BMPString, val=");
                        WriteString ("not yet implemented");  WriteLn;
                        Skip (shortlength, varlength);
                ELSE
                        WriteLn;  WriteString ("UNKNOWN TYPE ");
                        WriteCard (type);  WriteLn;
                        Skip (shortlength, varlength);
            END (*CASE*);
        END (*IF*);
        BigNum.Discard (length);

    END ParseASN1;

(************************************************************************)
(*                        TRANSLATE PEM TO BINARY                       *)
(************************************************************************)

PROCEDURE ReadPEM (VAR (*IN*) filename: FilenameString;
                                VAR (*OUT*) buffer: ARRAY OF CHAR);

    (* Loads a Base64 string from a file, deleting any line that    *)
    (* begins with the three characters '---'.                      *)

    VAR line: ARRAY [0..511] OF CHAR;
        cid: ChanId;

    BEGIN
        buffer[0] := Nul;
        cid := OpenOldFile (filename, FALSE, FALSE);
        LOOP
            ReadLine (cid, line);
            IF line[0] = CtrlZ THEN
                EXIT (*LOOP*);
            ELSIF (line[0] <> '-') AND (line[1] <> '-') AND (line[2] <> '-') THEN
                Strings.Append(line, buffer);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
    END ReadPEM;

(************************************************************************)

PROCEDURE PEMtoBinary (VAR (*IN*) infilename: FilenameString;
                        VAR (*OUT*) bindata: ARRAY OF CARD8): CARDINAL;

    (* Reads a PEM file, translates the data to binary in bindata.      *)
    (* Also writes bindata to an output file.  Returns the number of    *)
    (* output bytes.                                                    *)

    CONST
        InBufSize = 65536;

    VAR outfilename: FilenameString;
        inbuf: ARRAY [0..InBufSize-1] OF CHAR;
        N, pos: CARDINAL;  cid: ChanId;
        found: BOOLEAN;

    BEGIN
        (* Create the output file name from the input name. *)

        Strings.Assign (infilename, outfilename);
        Strings.FindPrev ('.', outfilename, LENGTH(outfilename) - 1,
                                                            found, pos);
        IF found THEN
            outfilename[pos] := Nul;
        END (*IF*);
        Strings.Append (".bin", outfilename);
        DeleteFile (outfilename);

        (* Do the translation. *)

        ReadPEM (infilename, inbuf);
        N := Decode (inbuf, bindata);
        IF N <> 0 THEN
            WriteString ("Creating file ");
            WriteString (outfilename);  WriteLn;
            cid := OpenNewFile (outfilename, TRUE);
            WriteRaw (cid, bindata, N);
            CloseFile (cid);
        END (*IF*);

        RETURN N;

    END PEMtoBinary;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoDecode;

    CONST BufSize = 65536;

    VAR infilename: FilenameString;
        binbuff: ARRAY [0..BufSize-1] OF CARD8;
        extension: ARRAY [0..15] OF CHAR;
        N, pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        (* Get the input file name. *)

        GetParameter (infilename);
        N := Strings.Length (infilename);
        WHILE (N > 0) AND (infilename[N-1] = ' ') DO
            DEC(N);
        END (*WHILE*);
        infilename[N] := Nul;
        IF infilename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        END (*IF*);

        (* What sort of file is it?  Does it exist?  *)

        IF NOT Exists (infilename) THEN
            Strings.Append (".bin", infilename);
        END (*IF*);
        Strings.FindPrev ('.', infilename, LENGTH(infilename) - 1,
                                                            found, pos);
        IF found THEN
            Strings.Extract (infilename, pos+1, SIZE(extension), extension);
        ELSE
            extension := "";
        END (*IF*);
        IF NOT Exists (infilename) THEN
            infilename[pos] := Nul;
            Strings.Append (".pem", infilename);
            extension := "PEM";
        END (*IF*);

        (* Do the decoding.  *)

        WriteString ("Reading from file ");
        WriteString (infilename);  WriteLn;
        IF StringMatch (extension, "PEM") THEN
            N := PEMtoBinary (infilename, binbuff);
        ELSE
            N := LoadFile (infilename, binbuff);
        END (*IF*);
        IF N = 0 THEN
            WriteString ("Could not load input file ");
            WriteString (infilename);
            WriteLn;
        ELSE
            (*WriteHexString (binbuff, 0, N);  WriteLn;*)
            WriteCard (N);  WriteString (" bytes total");  WriteLn;
            pos := 0;
            ParseASN1 (binbuff, pos, N, 0);
        END (*IF*);

    END DoDecode;

(************************************************************************)

BEGIN
    DoDecode;
END ASN1decode.

