(**************************************************************************)
(*                                                                        *)
(*  PMOS/2 software library                                               *)
(*  Copyright (C) 2015   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

(*
<* WOFF316+ *>
*)

IMPLEMENTATION MODULE RSA;

        (********************************************************)
        (*                                                      *)
        (*                    RSA encryption                    *)
        (*                                                      *)
        (*     Implementation of RSA as defined in RFC3447      *)
        (*                                                      *)
        (*    My goal is to implement rsa-sha1 and rsa-sha256   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 October 2015                 *)
        (*  Last edited:        29 October 2015                 *)
        (*  Status:             Just starting                   *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT
    (* type *)  CARD8, CARD32, ADDRESS,
    (* proc *)  ADR;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

(*
FROM MD5 IMPORT
    (* type *)  MD5_CTX, MD5_DigestType,
    (* proc *)  MD5Init, MD5Update, MD5Final, MD5DigestToString;

FROM SHA1 IMPORT
    (* type *)  SHA1_CTX, SHA1_DigestType,
    (* proc *)  SHA1Init, SHA1Update, SHA1Final, SHA1DigestToString;

FROM LowLevel IMPORT
    (* proc *)  IXORB;

FROM SYSTEM IMPORT
    (* type *)  CARD8, LOC,
    (* proc *)  ADR, MOVE, FILL;
*)

(************************************************************************)
(*                                                                      *)
(*  REMARK: There is a certain amount of code duplication below.  I did *)
(*  look at the possibility of having a single generic HMAC procedure,  *)
(*  but in the final analysis it seemed better to accept the redundancy *)
(*  for the sake of better run-time efficiency.                         *)
(*                                                                      *)
(************************************************************************)

TYPE
    Card5 = ARRAY [0..4] OF CARD32;
    Card8 = ARRAY [0..7] OF CARD32;

VAR
    SHA1_hash: Card5;
    SHA256_hash: Card8;

(************************************************************************)
(*                           SUPPORT ROUTINES                           *)
(************************************************************************)

(************************************************************************)
(*                          MAIN RSA ALGORITHM                          *)
(************************************************************************)

(*   SOME IMPORTANT DEFINITIONS
        k length in octets of the RSA modulus n
        mLen length in octets of a message M
        n RSA modulus, n = r_1 * r_2 * ... * r_u , u >= 2
        (n, e) RSA public key
        p, q first two prime factors of the RSA modulus n
*)

(************************************************************************)

PROCEDURE EME_OAEP_encode (option, k: CARDINAL);

    (* EME-OAEP encoding; See Figure 1 in the RFC.  *)

    CONST L = "";
         DBsize = 1024;

    VAR i, j, hLen, size: CARDINAL;
        src: ADDRESS;
        DB: ARRAY [0..DBsize-1] OF CARD8;

    BEGIN
        (*                                                              *)
        (*    a. If the label L is not provided, let L be the empty string. Let *)
        (*       lHash = Hash(L), an octet string of length hLen (see the note *)
        (*       below).                                                          *)

        IF option = 1 THEN
            hLen := 5*4;
            src := ADR(SHA1_hash);
        ELSIF option = 256 THEN
            hLen := 8*4;
            src := ADR(SHA256_hash);
        ELSE
            WriteString ("ERROR: unsupported encoding method");  WriteLn;
            RETURN;
        END (*IF*);
        Copy (src, ADR(DB), hLen);
        j := hLen;

        (*                                                                      *)
        (*    b. Generate an octet string PS consisting of k - mLen - 2hLen - 2  *)
        (*       zero octets.  The length of PS may be zero.                     *)

        size := k - mLen - 2*hLen - 2;
        IF size > 0 THEN
            FOR i := 0 TO size-1 DO
                DB[j] := 0;  INC(j);
            END (*FOR*);
        END (*IF*);

        (*                                                                      *)
        (*    c. Concatenate lHash, PS, a single octet with hexadecimal value   *)
        (*       0x01, and the message M to form a data block DB of length k -   *)
        (*       hLen - 1 octets as                                             *)
        (*                                                                      *)
        (*          DB = lHash || PS || 0x01 || M.                             *)
        (*                                                                      *)

        DB[j] := 1;  INC(j);

        (*    d. Generate a random octet string seed of length hLen.            *)
        (*                                                                      *)
        (*    e. Let dbMask = MGF(seed, k - hLen - 1).                          *)
        (*                                                                      *)
        (*    f. Let maskedDB = DB \xor dbMask.                                 *)
        (*                                                                      *)
        (*    g. Let seedMask = MGF(maskedDB, hLen).                            *)
        (*                                                                      *)
        (*    h. Let maskedSeed = seed \xor seedMask.                           *)
        (*                                                                      *)
        (*    i. Concatenate a single octet with hexadecimal value 0x00,        *)
        (*       maskedSeed, and maskedDB to form an encoded message EM of     *)
        (*       length k octets as                                           *)
        (*                                                                    *)
        (*          EM = 0x00 || maskedSeed || maskedDB.                      *)

    END EME_OAEP_encode;

(************************************************************************)

PROCEDURE RSAEncrypt (option, k: CARDINAL);

    BEGIN
        (* Section numbers in the comments refer to RFC 3447.   *)

        (* 1. Length checking:                                         *)
        (*                                                               *)
        (*    a. If the length of L is greater than the input limitation for the *)
        (*       hash function (2^61 - 1 octets for SHA-1), output "label too    *)
        (*       long" and stop.                                              *)
        (*       (In our case, L is the null string. *)
        (*                                                                   *)
        (*    b. If mLen > k - 2hLen - 2, output "message too long" and stop.  *)

        (* 2. EME-OAEP encoding (see Figure 1 below):   *)

        EME_OAEP_encode (option, k);

        (* a. Convert the encoded message EM to an integer message  *)
        (*    representative m (see Section 4.2):                   *)
        (*                                                          *)
        (*    m = OS2IP (EM).                                       *)

        (* b. Apply the RSAEP encryption primitive (Section 5.1.1) to   *)
        (*    the RSA public key (n, e) and the message representative  *)
        (*    m to produce an integer ciphertext representative c:      *)
        (*                                                              *)
        (*    c = RSAEP ((n, e), m).                                    *)

        (* c. Convert the ciphertext representative c to a ciphertext   *)
        (*    C of length k octets (see Section 4.1):                   *)
        (*                                                              *)
        (*    C = I2OSP (c, k).                                         *)

    END RSAEncrypt;

(************************************************************************)

BEGIN
     SHA1_hash := Card5 {0DA39A3EEH, 5E6B4B0DH, 3255BFEFH, 95601890H, 0AFD80709H};
     SHA256_hash := Card8 {0E3B0C442H, 98FC1C14H, 9AFBF4C8H, 996FB924H, 27AE41E4H,
                            649B934CH, 0A495991BH, 7852B855H};

END RSA.

