(**************************************************************************)
(*                                                                        *)
(*  Web server input stream handler                                       *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLSMAC;

        (********************************************************)
        (*                                                      *)
        (*  Interface between TLS and the encryption algorithms *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            19 July 2018                    *)
        (*  Last edited:        18 July 2018                    *)
        (*  Status:             Just starting                   *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, CARD16;

IMPORT Strings;

FROM STextIO IMPORT                 (* for debugging *)
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM TLSBase IMPORT
    (* type *)  FragmentType,
    (* proc *)  NYI;

(*
FROM SYSTEM IMPORT CARD8, CARD16, ADR;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  recv, soclose;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;
*)

(************************************************************************)

TYPE
    BulkCipherAlgorithm = (nocipher, rc4_128, tripledes, aes128, aes256);

    EncrRule = POINTER TO
                    RECORD
                        bulkcipher: BulkCipherAlgorithm;
                    END (*RECORD*);

    PRFAlgorithm = (tls_prf_sha256);
    KeyExchangeAlgorithm = (nullkeyx, RSA);
    CipherType = (stream, block, aead);
    MACAlgorithm = (null, hmac_md5, hmac_sha1, hmac_sha256, hmac_sha384,
                                 hmac_sha512);

    (* See RFC 5246, Appendix C, for how these decompose into three     *)
    (* critical properties: key exchange, bulk cipher, and MAC.  There  *)
    (* is obviously extensive use of suites that are not mentioned in   *)
    (* RFC 5246, and for those I am relying on web sources.             *)

    SuitePropertyRecord =
            RECORD
                supported       : BOOLEAN;
                code1, code2    : CARD8;
                keyx            : KeyExchangeAlgorithm;
                cipher          : BulkCipherAlgorithm;
                mac             : MACAlgorithm;
            END (*RECORD*);

    PropertyList = ARRAY CipherSuite OF SuitePropertyRecord;

CONST
    SuiteProperty = PropertyList
            {
            {FALSE, 0, 0, nullkeyx, nocipher, null},
            {FALSE, 0, 1, RSA, nocipher, hmac_md5},
            {FALSE, 0, 2, RSA, nocipher, hmac_sha1},
            {FALSE, 0, 3BH, RSA, nocipher, hmac_sha256},
            {FALSE, 0, 4, RSA, rc4_128, hmac_md5},
            {FALSE, 0, 5, RSA, rc4_128, hmac_sha1},
            {FALSE, 0, 0AH, RSA, tripledes, hmac_sha1},
            {TRUE,  0, 2FH, RSA, aes128, hmac_sha1},
            {FALSE, 0, 35H, RSA, aes256, hmac_sha1},
            {FALSE, 0, 3CH, RSA, aes128, hmac_sha256},
            {FALSE, 0, 3DH, RSA, aes256, hmac_sha256}
            };

    (* The list of suites I get from OS/2 Firefox, in preference order, is
        00 1E       TLS_KRB5_WITH_DES_CBC_SHA (not implemented)
        C0 2B       TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256 (not implemented)
        C0 2F       TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 (not implemented)
        C0 0A       TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA (not implemented)
        C0 09       TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA (not implemented)
        C0 13       TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA (not implemented)
        C0 14       TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA (not implemented)
        C0 07       TLS_ECDHE_ECDSA_WITH_RC4_128_SHA (not implemented)
        C0 11       TLS_ECDHE_RSA_WITH_RC4_128_SHA (not implemented)
        00 33       TLS_DHE_RSA_WITH_AES_128_CBC_SHA (not implemented)
        00 39       TLS_DHE_RSA_WITH_AES_256_CBC_SHA (not implemented)
        00 2F       TLS_RSA_WITH_AES_128_CBC_SHA (done)
        00 35       TLS_RSA_WITH_AES_256_CBC_SHA (not implemented, but I should consider this one)
        00 0A       TLS_RSA_WITH_3DES_EDE_CBC_SHA

    Clearly there is a disconnect between what I have implemented and
    what Firefox prefers.  To be compatible, I need to do one or more
    of the following:
      - implement the CBC version of AES; DONE;
      - implement RC4 - DONE;
      - look into ECDHE - probably hard, but worth a look.
      - implement 3DES - DONE.
    *)

(*
    Byte32 = ARRAY [0..31] OF CARD8;
    Byte48 = ARRAY [0..47] OF CARD8;

CONST
    Zero32 = Byte32 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    Zero48 = Byte48 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

TYPE
    TLSsession = POINTER TO SessionState;

    ConnectionEnd = (server, client);
    PRFAlgorithm = (tls_prf_sha256);
    BulkCipherAlgorithm = (nocipher, rc4, tripledes, aes);
    CipherType = (stream, block, aead);
    MACAlgorithm = (null, hmac_md5, hmac_sha1, hmac_sha256, hmac_sha384,
                                 hmac_sha512);
    CompressionMethod = (nocompression, other);

    SecurityParameters =
            RECORD
                entity:                 ConnectionEnd;     (* constant? *)
                prf_algorithm:          PRFAlgorithm;      (* constant? *)
                bulk_cipher_algorithm:  BulkCipherAlgorithm;  (* used? *)
                cipher_type:            CipherType;        (* used? *)
                enc_key_length:         CARD8;             (* used? *)
                block_length:           CARD8;             (* used? *)
                fixed_iv_length:        CARD8;             (* used? *)
                record_iv_length:       CARD8;             (* used? *)
                mac_algorithm:          MACAlgorithm;      (* used? *)
                mac_length:             CARD8;             (* used? *)
                mac_key_length:         CARD8;             (* used? *)
                compression_algorithm:  CompressionMethod; (* used? *)
                master_secret:          Byte48;            (* used? *)
                client_random:          Byte32;            (* used? *)
                server_random:          Byte32;            (* used? *)
            END (*RECORD*);

    SessionState =  RECORD
                        sock:   Socket;
                        security: SecurityParameters;
                    END (*RECORD*);

    FragmentType = POINTER TO ARRAY [0..65535] OF CARD8;

    (* The standard defines a type "opaque" for both fixed-length and   *)
    (* variable-length arrays.  In the variable-length case only, such  *)
    (* an array is passed in messages as a length followed by the data. *)
    (* The number of bytes in the length depends on the maximum array   *)
    (* length; we therefore define two types, "opaque short" and        *)
    (* "opaque medium".  These are used only for variable-length byte   *)
    (* strings.  For fixed-length, we just use a fixed-length data type.*)

    OpaqueS =   RECORD
                    length: CARD8;
                    string: FragmentType;
                END (*RECORD*);

    OpaqueM =   RECORD
                    length: CARD16;
                    string: FragmentType;
                END (*RECORD*);

*)
(************************************************************************)

PROCEDURE NameOfSuite (s: CipherSuite;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns name as the name of ciphersuite s. *)

    TYPE SNtype = ARRAY CipherSuite OF ARRAY [0..32] OF CHAR;

    CONST
        SuiteName = SNtype {"TLS_NULL_WITH_NULL_NULL",
                            "TLS_RSA_WITH_NULL_MD5",
                            "TLS_RSA_WITH_NULL_SHA",
                            "TLS_RSA_WITH_NULL_SHA256",
                            "TLS_RSA_WITH_RC4_128_MD5",
                            "TLS_RSA_WITH_RC4_128_SHA",
                            "TLS_RSA_WITH_3DES_EDE_CBC_SHA",
                            "TLS_RSA_WITH_AES_128_CBC_SHA",
                            "TLS_RSA_WITH_AES_256_CBC_SHA",
                            "TLS_RSA_WITH_AES_128_CBC_SHA256",
                            "TLS_RSA_WITH_AES_256_CBC_SHA256"
                            };

    BEGIN
        Strings.Assign (SuiteName[s], name);
    END NameOfSuite;

(************************************************************************)
(*                       ACCEPTING A CIPHER SUITE                       *)
(************************************************************************)

PROCEDURE AcceptCipherSuite (code1, code2: CARD8): CipherSuite;

    (* Returns result <> NullSuite iff we support this cipher suite. *)

    VAR j: CipherSuite;

    BEGIN
        j := NullSuite;
        WHILE j < MAX(CipherSuite) DO
            INC (j);
            IF (SuiteProperty[j].code1 = code1)
                    AND (SuiteProperty[j].code2 = code2)
                    AND SuiteProperty[j].supported THEN
                RETURN j;
            END (*IF*);
        END (*LOOP*);
        RETURN NullSuite;
    END AcceptCipherSuite;

(************************************************************************)
(*                               ENCRYPTION                             *)
(************************************************************************)

PROCEDURE Encrypt (rule: EncrRule;  VAR (*INOUT*) message: FragmentType;  length: CARDINAL)
                                                               : CARDINAL;

    (* The message is replaced by its encrypted version.  The returned  *)
    (* value is the new length.                                         *)

    BEGIN
        IF (rule <> NIL) AND (rule^.bulkcipher <> nocipher) THEN
            NYI ("Encrypt");
        END (*IF*);
        RETURN length;
    END Encrypt;

(************************************************************************)

PROCEDURE MACgen (rule: EncrRule;  VAR (*INOUT*) message: FragmentType;  length: CARDINAL)
                                                               : CARDINAL;

    (* The message is replaced by its encrypted version.  The returned  *)
    (* value is the new length.                                         *)

    BEGIN
        MACIni();
        MACcontinue (seq_num);
        MACcontinue (type);
        MACcontinue (versi0on);
        MACcontinue (length);
        MACcontinue (fragment);
    END Encrypt;

(************************************************************************)

END TLSMAC.

