(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

<* WOFF316+ *>

IMPLEMENTATION MODULE TLSnotes;

        (********************************************************)
        (*                                                      *)
        (*   Miscellaneous informatin collected from RFC 5246   *)
        (*                                                      *)
        (*     At present this module contains nothing but      *)
        (*     pseudocode.                                      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            30 March 2018                   *)
        (*  Last edited:        30 March 2018                   *)
        (*  Status:             Just a skeleton for now         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT
    (* type *)  CARD8, CARD16, ADDRESS;

FROM Arith64 IMPORT
    (* type *)  CARD64;

(************************************************************************)
(*                                                                      *)
(*                         IMPORTANT DATA TYPES                         *)
(*                                                                      *)
(*  Taken from RFC 5246.                                                *)
(*                                                                      *)
(************************************************************************)

TYPE
    (* The standard uses this type repeatedly to mean a variable-length *)
    (* string of bytes.  I have chosen the obvious implementation.      *)

    Opaque = RECORD
                length: CARDINAL;
                content: ADDRESS;
            END (*RECORD*);

    DigitallySigned =
            RECORD
                algorithm: SignatureAndHashAlgorithm;  (* section 7.4.1.4.1 *)
                signature: Opaque;
            END (*RECORD*);

    ConnectionEnd = (server, client);
    PRFAlgorithm = (tls_prf_sha256);
    BulkCipherAlgorithm = (nocipher, rc4, tripledes, aes);
    CipherType = (stream, block, aead);
    CompressionMethod = (nocompression, other);

    SecurityParameters =
            RECORD
                entity:             ConnectionEnd;
                prf_algorithm:      PRFAlgorithm;
                bulk_cipher_algorithm: BulkCipherAlgorithm;
                cipher_type:        CipherType;
                enc_key_length:     CARD8;
                block_length:       CARD8;
                fixed_iv_length:    CARD8;
                record_iv_length:   CARD8;
                mac_algorithm:      MACAlgorithm;
                mac_length:         CARD8;
                mac_key_length:     CARD8;
                compression_algorithm:  CompressionMethod;
                master_secret:      ARRAY [0..47] OF CARD8;
                client_random:      ARRAY [0..31] OF CARD8;
                server_random:      ARRAY [0..31] OF CARD8;
            END (*RECORD*);

    ConnectionState =
            RECORD
                compressionstate: CARD8;
                cipherstate:    CARD8;
                MACkey:     CARD8;
                readseqnum: CARD64;
                writeseqnum:    CARD64;
            END (*RECORD*);

    (* Record layer types. *)

    ProtocolVersion =
            RECORD
                major, minor: CARD8;
            END (*RECORD*);

    ContentType = CARD8;

CONST
    change_cipher_spec = VAL (ContentType, 20);
    alert = VAL (ContentType, 21);
    handshake = VAL (ContentType, 22);
    application_data = VAL (ContentType, 23);

TYPE
    TLSPlaintext =
            RECORD
                type: ContentType;
                version: ProtocolVersion;
                length: CARD16;         (* limit 2^14 *)
                fragment: Opaque;       (* redundant length field here *)
            END (*RECORD*);

    TLSCompressed =
            RECORD                  (* apparently identical to TLSPlaintext *)
                type: ContentType;
                version: ProtocolVersion;
                length: CARD16;
                fragment: Opaque;
            END (*RECORD*);

    GenericStreamCipher =
            RECORD
                content: Opaque;     (* length is TLSCompressed.length *)
                MAC:    Opaque;         (* length is SecurityParameters.mac_length *)
                    (* see section 6.2.3.1 for MAC generation *)
            END (*RECORD*);

    GenericBlockCipher =
            RECORD
                IV: Opaque;         (* length is SecurityParameters.record_iv_length *)
                block_ciphered:
                    RECORD
                        content: Opaque;     (* length is TLSCompressed.length *)
                        MAC:    Opaque;         (* length is SecurityParameters.mac_length *)
                        padding: Opaque;    (* really should be array of CARD8 *)
                        padding_length: CARD8;
                        (* see section 6.2.3.2 for IV and padding. *)
                    END (*RECORD*);
            END (*RECORD*);

    GenericAEADCipher =
            RECORD
                (* Details omitted because I don't plan to include  *)
                (* any AEAD ciphers for now.                        *)
            END (*RECORD *);

    TLSCiphertext =
            RECORD
                type: ContentType;
                version: ProtocolVersion;
                length: CARD16;
                CASE fragment: CipherType OF
                    stream:     cs: GenericStreamCipher;
                    |
                    block:      cb: GenericBlockCipher;
                    |
                    aead:       ca: GenericAEADCipher;
                END (*CASE*);
            END (*RECORD*);

    (* We're now up to page 22 of RFC 5246. *)
    (* The block or stream cipher converts TLSCompressed.fragment       *)
    (* structures to and from TLSCiphertext.fragment structures.        *)

    (* TLS handshaking protocol codes.  One-byte message unless     *)
    (* otherwise specified.                                         *)

CONST
    ChangeCipherSpec = 1;

TYPE
    AlertLevel = CARD8;     (* 1 = warning, 2 = fatal *)
    AlertDescription = CARD8;
        (* codes are:
                  close_notify(0),
                  unexpected_message(10),
                  bad_record_mac(20),
                  decryption_failed_RESERVED(21),
                  record_overflow(22),
                  decompression_failure(30),
                  handshake_failure(40),
                  no_certificate_RESERVED(41),
                  bad_certificate(42),
                  unsupported_certificate(43),
                  certificate_revoked(44),
                  certificate_expired(45),
                  certificate_unknown(46),
                  illegal_parameter(47),
                  unknown_ca(48),
                  access_denied(49),
                  decode_error(50),
                  decrypt_error(51),
                  export_restriction_RESERVED(60),
                  protocol_version(70),
                  insufficient_security(71),
                  internal_error(80),
                  user_canceled(90),
                  no_renegotiation(100),
                  unsupported_extension(110)   *)

        (* See section 7.2.2 for more detail on alerts. *)

    Alert = RECORD
                level: AlertLevel;
                description: AlertDescription;
            END (*RECORD *);

    (* Now up to page 37, Handshake protocol, which might need a    *)
    (* new module.                                                  *)

(************************************************************************)

PROCEDURE GenerateMAC;

    (* MAC means Message Authentication code.   *)
    (* This is for a stream cipher, see 6.2.3.1. *)

    BEGIN
        (*
        The MAC is generated as:
        MAC(MAC_write_key, seq_num +
                           TLSCompressed.type +
                           TLSCompressed.version +
                           TLSCompressed.length +
                           TLSCompressed.fragment);
        where "+" denotes concatenation.
        seq_num
            The sequence number for this record.
        MAC
            The MAC algorithm specified by SecurityParameters.mac_algorithm.

        Note that the MAC is computed before encryption. A stream
        cipher encrypts the entire block, including the MAC, and it
        looks as if this is also true for block ciphers.
        *)
    END GenerateMAC;

(************************************************************************)

PROCEDURE dummy1;

    (* Dummy replacement for an external procedure. *)

    BEGIN
    END dummy1;

(************************************************************************)

PROCEDURE dummy2;

    BEGIN
    END dummy2;

(************************************************************************)


(************************************************************************)

END TLSnotes.

