(**************************************************************************)
(*                                                                        *)
(*  Translating a PEM (Privacy Enhanced Mail) file to binary              *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE PEM2bin;

        (********************************************************)
        (*                                                      *)
        (*      Translating a PEM (Privacy Enhanced Mail)       *)
        (*                    file to binary                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            27 June 2018                    *)
        (*  Last edited:        27 June 2018                    *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, TextIO, IOChan, BigNum;

FROM SYSTEM IMPORT CARD8;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  DeleteFile, OpenOldFile, OpenNewFile, CloseFile,
                ReadLine, WriteRaw;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Base64 IMPORT
    (* proc *)  Decode;

(************************************************************************)

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = "ex2";

(************************************************************************)
(*                     MISCELLANOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE WriteHexChar (N: CARD8);

    (* Write a single hexadecimal digit.  *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(ORD('A') + (N - 10)));
        ELSE
            WriteChar (CHR(ORD('0') + N));
        END (*IF*);
    END WriteHexChar;

(************************************************************************)

PROCEDURE WriteHex (N: CARD8);

    (* Write N in hexadecimal.  *)

    BEGIN
        WriteHexChar (N DIV 16);
        WriteHexChar (N MOD 16);
    END WriteHex;

(************************************************************************)

PROCEDURE WriteHexString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Write N bytes starting at buf[pos].  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteHex (buf[pos]);  INC(pos);
        END (*FOR*);
    END WriteHexString;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;
        N: CARDINAL;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);

        (* Remove leading and trailing spaces. *)

        WHILE result[0] = ' ' DO
            Strings.Delete (result, 0, 1);
        END (*WHILE*);
        N := Strings.Length(result);
        WHILE (N > 0) AND (result[N-1] = ' ') DO
            DEC (N);
        END (*WHILE*);
        result[N] := Nul;

    END GetParameter;

(************************************************************************)

PROCEDURE ReadPEM (filename: ARRAY OF CHAR;  VAR (*OUT*) buffer: ARRAY OF CHAR);

    (* Loads a Base64 string from a file. *)

    VAR line: ARRAY [0..511] OF CHAR;
        cid: ChanId;

    BEGIN
        buffer[0] := Nul;
        cid := OpenOldFile (filename, FALSE, FALSE);
        LOOP
            ReadLine (cid, line);
            IF line[0] = CtrlZ THEN
                EXIT (*LOOP*);
            ELSE
                Strings.Append(line, buffer);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
    END ReadPEM;

(************************************************************************)
(*                    WRITING THE BINARY OUTPUT FILE                    *)
(************************************************************************)

PROCEDURE WriteBinary (outname: ARRAY OF CHAR;
                        VAR (*IN*) binbuff: ARRAY OF CARD8;  N: CARDINAL);

    (* Writes N bytes of binbuff to file outname. *)

    VAR cid: ChanId;

    BEGIN
        DeleteFile (outname);
        cid := OpenNewFile (outname, TRUE);
        WriteRaw (cid, binbuff, N);
        CloseFile (cid);
    END WriteBinary;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoTranslate;

    (* For now, it appears to be sufficient to do an in-memory operation. *)

    CONST
        InBufSize = 65536;
        OutBufSize = 65536;

    VAR infilename, outfilename: ARRAY [0..511] OF CHAR;
        inbuf: ARRAY [0..InBufSize-1] OF CHAR;
        binbuf: ARRAY [0..OutBufSize-1] OF CARD8;
        N, pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        GetParameter (infilename);
        IF infilename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        ELSE
            WriteString ("infilename = ");  WriteString (infilename);
            WriteLn;
        END (*IF*);

        (* Create the output file name from the input name. *)

        Strings.Assign (infilename, outfilename);
        Strings.FindPrev ('.', outfilename, LENGTH(outfilename) - 1,
                                                            found, pos);
        IF found THEN
            outfilename[pos] := Nul;
        ELSE
            Strings.Append (".pem", infilename);
        END (*IF*);
            WriteString ("infilename = ");  WriteString (infilename);
            WriteLn;
        Strings.Append (".bin", outfilename);

        (* Do the translation. *)

        ReadPEM (infilename, inbuf);
        N := Decode (inbuf, binbuf);
        IF N = 0 THEN
            WriteString ("ERROR: No input data");
        ELSE
            WriteHexString (binbuf, 0, N);  WriteLn;
            WriteCard (N);  WriteString (" bytes total");  WriteLn;
            WriteBinary (outfilename, binbuf, N);
        END (*IF*);

    END DoTranslate;

(************************************************************************)

BEGIN
    DoTranslate;
END PEM2bin.

