(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

<* WOFF316+ *>

IMPLEMENTATION MODULE TLSPRF;

        (********************************************************)
        (*                                                      *)
        (*      Pseudo-random function for key generation       *)
        (*         as specified in RFC 5246, section 5.         *)
        (*                                                      *)
        (*     At present this module contains nothing but      *)
        (*     pseudocode.                                      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 October 2017                  *)
        (*  Last edited:        1 April 2018                    *)
        (*  Status:             Just a skeleton for now         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT
    (* type *)  LOC, ADDRESS,
    (* proc *)  ADR;

FROM LowLevel IMPORT
    (* proc *)  Copy, AddOffset;

FROM SHA2 IMPORT
    (* type *)  SHA2_DigestType;

FROM HMAC IMPORT
    (* proc *)  HMAC_SHA256;

(************************************************************************)
(*                                                                      *)
(*                          GENERIC FUNCTIONS                           *)
(*                                                                      *)
(*  This section contains mostly pseudo-code, as a template for future  *)
(*  expansion into real code.                                           *)
(*                                                                      *)
(************************************************************************)

TYPE
    (* The standard is vague about the types of the parameters, so I am *)
    (* defaulting to the SHA_256 case.                                  *)

    DigestType = SHA2_DigestType;
    DigestString = ARRAY [0..31] OF LOC;

(************************************************************************)

PROCEDURE HMAC_hash (secret: ARRAY OF LOC;  secretlength: CARDINAL;
                         input: ARRAY OF LOC;  inputlength: CARDINAL;
                            VAR (*OUT*) Digest: DigestType);

    (* This is for the specific case of HMAC_SHA256, which is what      *)
    (* RFC 5246 says that cipher suites SHOULD use, unless they choose  *)
    (* to use a stronger standard hash function.                        *)

    BEGIN
        HMAC_SHA256 (secret, secretlength, input, inputlength, Digest);
    END HMAC_hash;

(************************************************************************)

PROCEDURE P_hash (secretlength: CARDINAL;  secret: ARRAY OF LOC;
                  seedlength: CARDINAL;  seed: DigestType;
                  resultbytes: CARDINAL;  VAR (*OUT*) result: ARRAY OF LOC);

    (* Produces a pseudorandom result of length resultbytes. *)

    (* More work needed on seed type, because procedure PRF below calls *)
    (* this procedure with a seed that's longer than SIZE(DigestType).  *)
    (* We need two data types for A, one as a Digest and one as a       *)
    (* string, so one extra function ConvertToString (A).               *)

    VAR A, nextpart: DigestType;
        p: ADDRESS;
        pos, needed, partlength: CARDINAL;

    BEGIN
        needed := resultbytes;
        pos := 0;
        p := ADR(result);
        A := seed;
        REPEAT
            (* Probably have to swap endianness of A at this point.     *)
            (* Note that HMAC inherits the endianness of its hash       *)
            (* function, and the SHA functions all produce BigEndian    *)
            (* results.                                                 *)

            HMAC_hash (secret, secretlength, A, SIZE(A), A);

            (* And swap endianness yet again here while constructing    *)
            (* the string (secret+A).                                   *)

            HMAC_hash (secret, secretlength, A (*followed by seed*),
                                            SIZE(A)+seedlength, nextpart);
            IF partlength > needed THEN
                partlength := needed;
            END (*IF*);
            Copy (ADR(nextpart), p, partlength);
            INC (pos, partlength);
            p := AddOffset (p, partlength);
            DEC (needed, partlength);
        UNTIL needed = 0;

    END P_hash;

(************************************************************************)

PROCEDURE PRF (secretlength: CARDINAL;  secret: ARRAY OF LOC;
                label: ARRAY OF CHAR;
                seedlength: CARDINAL;  seed: DigestType;
                resultbytes: CARDINAL;  VAR (*OUT*) result: ARRAY OF LOC);

    VAR newseed: ARRAY [0..2048] OF CHAR;

    BEGIN
        (* Let newseed be label followed by BigEndian seed converted    *)
        (* to a byte string.                                            *)

        P_hash (secretlength, secret, LENGTH(label) + SIZE(seed), newseed,
                  resultbytes, result);

    END PRF;

(************************************************************************)

PROCEDURE GenerateKeys;

    (* Section 6.3 of the TLS standard.  We use the PRF function to     *)
    (* generate a key string, then we partition it into the up to six   *)
    (* keys that we really want:                                        *)
    (*      client_write_MAC_key[SecurityParameters.mac_key_length]     *)
    (*      server_write_MAC_key[SecurityParameters.mac_key_length]     *)
    (*      client_write_key[SecurityParameters.enc_key_length]         *)
    (*      server_write_key[SecurityParameters.enc_key_length]         *)
    (*      client_write_IV[SecurityParameters.fixed_iv_length]         *)
    (*      server_write_IV[SecurityParameters.fixed_iv_length]         *)
    (* Unused values are empty.                                         *)

    (* Implementation note: The currently defined cipher suite which    *)
    (* requires the most material is AES_256_CBC_SHA256. It requires    *)
    (* 2 x 32 byte keys and 2 x 32 byte MAC keys, for a total 128 bytes *)
    (* of key material.                                                 *)

    BEGIN
        (*
        key_block = PRF(SecurityParameters.master_secret,
                           "key expansion",
                           SecurityParameters.server_random +
                           SecurityParameters.client_random);
        *)
    END GenerateKeys;

(************************************************************************)

END TLSPRF.

