(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)


MODULE AEStest;

        (********************************************************)
        (*                                                      *)
        (*    AES: Keyed hashing for message authentication     *)
        (*                                                      *)
        (*                Test of module AES                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 October 2017                 *)
        (*  Last edited:        4 February 2021                 *)
        (*  Status:             Passes all test cases in the    *)
        (*                      standard NIST.FIPS.197          *)
        (*                                                      *)
        (*                      For CBC, passes cases 1..4 in   *)
        (*                      RFC 3602.  I have not yet had   *)
        (*                      the patience to run the         *)
        (*                      remaining cases.                *)
        (*                                                      *)
        (*      For CTR, passes every test in RFC 3686.         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, ADR;

IMPORT Strings;

FROM AES IMPORT
    (* type *)  AEScontext,
    (* proc *)  AESopen, AES_CBCopen, AES_CTRopen, AESprocess, AESfinal;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM HexData IMPORT
    (* proc *)  HexToBytes, DumpByteArray;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N in decimal *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE DoCrypt (fullpad: BOOLEAN;  version: CARDINAL;  key, text: ARRAY OF CHAR);

    (* Does an encryption followed by a decryption. *)

    VAR M, N, length: CARDINAL;
        binkey: ARRAY [0..63] OF CARD8;
        bintext, result: ARRAY [0..63] OF CARD8;
        ctx: AEScontext;

    BEGIN
        M := HexToBytes (text, bintext);
        N := HexToBytes (key, binkey);
        WriteString (" key = ");  DumpByteArray (N, binkey);  WriteLn;
        WriteString ("text = ");  DumpByteArray (M, bintext);  WriteLn;
        ctx := AESopen (FALSE, fullpad, version, binkey);
        length := 0;
        AESprocess (ctx, M, bintext, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AESopen (TRUE, fullpad, version, binkey);
        M := 0;
        AESprocess (ctx, length, result, bintext, M);
        AESfinal (ctx, bintext, M);
        WriteString ("After reverse operation,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, bintext);  WriteLn;
        WriteLn;

    END DoCrypt;

(************************************************************************)

PROCEDURE DoCBCgen (version, Nk: CARDINAL;  key: ARRAY OF CARD8;
                             NI: CARDINAL;  IV: ARRAY OF CARD8;
                             Nt: CARDINAL;  text: ARRAY OF CARD8);

    (* Encryption followed by decryption, for the CBC variant of AES.   *)

    CONST fullpad = TRUE;

    VAR M, length: CARDINAL;
        result: ARRAY [0..255] OF CARD8;
        ctx: AEScontext;

    BEGIN
        WriteString (" key = ");  DumpByteArray (Nk, key);  WriteLn;
        WriteString ("  IV = ");  DumpByteArray (NI, IV);  WriteLn;
        WriteString ("text = ");  DumpByteArray (Nt, text);  WriteLn;
        ctx := AES_CBCopen (FALSE, fullpad, version, key, IV);
        length := 0;
        AESprocess (ctx, Nt, text, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AES_CBCopen (TRUE, fullpad, version, key, IV);
        M := 0;
        AESprocess (ctx, length, result, text, M);
        AESfinal (ctx, text, M);
        WriteString ("After decryption,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, text);  WriteLn;
        WriteLn;

    END DoCBCgen;

(************************************************************************)

PROCEDURE DoCBChex (version: CARDINAL;  key, IV, text: ARRAY OF CHAR);

    (* The case where key, IV, text are all hex coded. *)

    VAR Nk, NI, Nt: CARDINAL;
        binkey, binIV, bintext: ARRAY [0..255] OF CARD8;

    BEGIN
        Nk := HexToBytes (key, binkey);
        NI := HexToBytes (IV, binIV);
        Nt := HexToBytes (text, bintext);
        DoCBCgen (version, Nk, binkey, NI, binIV, Nt, bintext);
    END DoCBChex;

(************************************************************************)

PROCEDURE DoCBCtxt (version: CARDINAL;  key, IV, text: ARRAY OF CHAR);

    (* The case where the text is ASCII text, not hex coded.    *)
    (* The key and IV are still hex coded.                      *)

    VAR Nk, NI, Nt: CARDINAL;
        binkey, binIV, bintext: ARRAY [0..255] OF CARD8;

    BEGIN
        Nk := HexToBytes (key, binkey);
        NI := HexToBytes (IV, binIV);
        Nt := Strings.Length (text);
        Copy (ADR(text), ADR(bintext), Nt);
        DoCBCgen (version, Nk, binkey, NI, binIV, Nt, bintext);
    END DoCBCtxt;

(************************************************************************)

PROCEDURE DoCTRgen (testnum: CARDINAL;
                    version, Nk: CARDINAL;  key: ARRAY OF CARD8;
                             NI: CARDINAL;  IV: ARRAY OF CARD8;
                             Nn: CARDINAL;  nonce: ARRAY OF CARD8;
                             Nt: CARDINAL;  text: ARRAY OF CARD8);

    (* Encryption followed by decryption, for the CTR variant of AES.   *)

    CONST fullpad = TRUE;

    VAR M, length, noncew: CARDINAL;
        result: ARRAY [0..255] OF CARD8;
        ctx: AEScontext;

    BEGIN
        WriteString ("Test Vector #");  WriteCard (testnum);  WriteLn;
        WriteString (" key = ");  DumpByteArray (Nk, key);  WriteLn;
        WriteString ("  IV = ");  DumpByteArray (NI, IV);  WriteLn;
        WriteString ("nonce= ");  DumpByteArray (Nn, nonce);  WriteLn;
        WriteString ("text = ");  DumpByteArray (Nt, text);  WriteLn;
        Copy (ADR(nonce), ADR(noncew), 4);
        ctx := AES_CTRopen (version, key, noncew, IV);
        length := 0;
        AESprocess (ctx, Nt, text, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AES_CTRopen (version, key, noncew, IV);
        M := 0;
        AESprocess (ctx, length, result, text, M);
        AESfinal (ctx, text, M);
        WriteString ("After decryption,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, text);  WriteLn;
        WriteLn;

    END DoCTRgen;

(************************************************************************)

PROCEDURE DoCTRhex (testnum, version: CARDINAL;  key, IV, nonce, text: ARRAY OF CHAR);

    (* The case where key, IV, nonce, text are all hex coded. *)

    VAR Nk, NI, Nn, Nt: CARDINAL;
        binkey, binIV, binnonce, bintext: ARRAY [0..255] OF CARD8;

    BEGIN
        Nk := HexToBytes (key, binkey);
        NI := HexToBytes (IV, binIV);
        Nn := HexToBytes (nonce, binnonce);
        Nt := HexToBytes (text, bintext);
        DoCTRgen (testnum, version, Nk, binkey, NI, binIV, Nn, binnonce, Nt, bintext);
    END DoCTRhex;

(************************************************************************)

PROCEDURE DribbleInputTest (version: CARDINAL;  key, text: ARRAY OF CHAR);

    (* Like DoCrypt, but we are checking whether we can safely pass     *)
    (* the data a little bit at a time.                                 *)

    CONST fullpad = FALSE;

    VAR M, N, length, j, L1, L2: CARDINAL;
        binkey: ARRAY [0..63] OF CARD8;
        bintext, result, part2: ARRAY [0..63] OF CARD8;
        ctx: AEScontext;

    BEGIN
        M := HexToBytes (text, bintext);
        N := HexToBytes (key, binkey);
        WriteString (" key = ");  DumpByteArray (N, binkey);  WriteLn;
        WriteString ("text = ");  DumpByteArray (M, bintext);  WriteLn;

        (* Now chop bintext into parts. *)

        L1 := M DIV 3;  L2 := M - L1;
        FOR j := 0 TO L2-1 DO
            part2[j] := bintext[j+L1];
        END (*FOR*);

        ctx := AESopen (FALSE, fullpad, version, binkey);
        length := 0;
        AESprocess (ctx, L1, bintext, result, length);
        AESprocess (ctx, L2, part2, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AESopen (TRUE, fullpad, version, binkey);
        M := 0;
        AESprocess (ctx, length, result, bintext, M);
        AESfinal (ctx, bintext, M);
        WriteString ("After decryption,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, bintext);  WriteLn;
        WriteLn;

    END DribbleInputTest;

(************************************************************************)

PROCEDURE BasicTest;

    (* Run some test vectors. *)

    CONST fullpad = FALSE;

    BEGIN
        DoCrypt (fullpad, 128, "2b 7e 15 16 28 ae d2 a6 ab f7 15 88 09 cf 4f 3c",
                            "32 43 f6 a8 88 5a 30 8d 31 31 98 a2 e0 37 07 34");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                            "00112233445566778899aabbccddeeff");
        DoCrypt (fullpad, 192, "000102030405060708090a0b0c0d0e0f1011121314151617",
                            "00112233445566778899aabbccddeeff");
        DoCrypt (fullpad, 256, "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",
                            "00112233445566778899aabbccddeeff");
        DribbleInputTest (128, "2b 7e 15 16 28 ae d2 a6 ab f7 15 88 09 cf 4f 3c",
                            "32 43 f6 a8 88 5a 30 8d 31 31 98 a2 e0 37 07 34");
    END BasicTest;

(************************************************************************)

PROCEDURE CBCtests;

    (* Test vectors taken from RFC 3602. *)

    BEGIN
        DoCBCtxt (128, "06a9214036b8a15b512e03d534120006",
                 "3dafba429d9eb430b422da802c9fac41", "Single block msg");
        DoCBChex (128, "c286696d887c9aa0611bbb3e2025a45a",
                 "562e17996d093d28ddb3ba695a2e6f58",
                  "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f");
        DoCBCtxt (128, "6c3ea0477630ce21a2ce334aa746c2cd",
                 "c782dc4c098c66cbd9cd27d825682c81", "This is a 48-byte message (exactly 3 AES blocks)");
        DoCBChex (128, "56e47a38c5598974bc46903dba290349",
                 "8ce82eefbea0da3c44699ed7db51b7d9",
                  "a0a1a2a3a4a5a6a7a8a9aaabacadaeaf b0b1b2b3b4b5b6b7b8b9babbbcbdbebf c0c1c2c3c4c5c6c7c8c9cacbcccdcecf d0d1d2d3d4d5d6d7d8d9dadbdcdddedf");
    END CBCtests;

(************************************************************************)

PROCEDURE CTRtests;

    (* Test vectors taken from RFC 3686. *)

    BEGIN
        DoCTRhex (1, 128, "AE 68 52 F8 12 10 67 CC 4B F7 A5 76 55 77 F3 9E",
                    "00 00 00 00 00 00 00 00", "00 00 00 30",
                     "53 69 6E 67 6C 65 20 62 6C 6F 63 6B 20 6D 73 67");
        DoCTRhex (2, 128, "7E 24 06 78 17 FA E0 D7 43 D6 CE 1F 32 53 91 63",
                    "C0 54 3B 59 DA 48 D9 0B", "00 6C B6 DB",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                        "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F");
        DoCTRhex (3, 128, "76 91 BE 03 5E 50 20 A8 AC 6E 61 85 29 F9 A0 DC",
                    "27 77 7F 3F  4A 17 86 F0", "00 E0 01 7B",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                    "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F 20 21 22 23");
        DoCTRhex (4, 192, "16 AF 5B 14 5F C9 F5 79 C1 75 F9 3E 3B FB 0E ED" +
                          "86 3D 06 CC FD B7 85 15",
                    "36 73 3C 14 7D 6D 93 CB", "00 00 00 48",
                     "53 69 6E 67 6C 65 20 62 6C 6F 63 6B 20 6D 73 67");
        DoCTRhex (5, 192, "7C 5C B2 40 1B 3D C3 3C 19 E7 34 08 19 E0 F6 9C" +
                         "67 8C 3D B8 E6 F6 A9 1A",
                    "02 0C 6E AD C2 CB 50 0D", "00 96 B0 3B",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                      "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F");
        DoCTRhex (6, 192, "02 BF 39 1E E8 EC B1 59 B9 59 61 7B 09 65 27 9B" +
                      "F5 9B 60 A7 86 D3 E0 FE",
                    "5C BD 60 27 8D CC 09 12", "00 07 BD FD",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                        "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F" +
                        "20 21 22 23");
        DoCTRhex (7, 256, "77 6B EF F2 85 1D B0 6F 4C 8A 05 42 C8 69 6F 6C" +
                        "6A 81 AF 1E EC 96 B4 D3 7F C1 D6 89 E6 C1 C1 04",
                    "DB 56 72 C9 7A A8 F0 B2", "00 00 00 60",
                     "53 69 6E 67 6C 65 20 62 6C 6F 63 6B 20 6D 73 67");
        DoCTRhex (8, 256, "F6 D6 6D 6B D5 2D 59 BB 07 96 36 58 79 EF F8 86" +
                        "C6 6D D5 1A 5B 6A 99 74 4B 50 59 0C 87 A2 38 84",
                    "C1 58 5E F1 5A 43 D8 75", "00 FA AC 24",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                        "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F");
        DoCTRhex (9, 256, "FF 7A 61 7C E6 91 48 E4 F1 72 6E 2F 43 58 1D E2" +
                        "AA 62 D9 F8 05 53 2E DF F1 EE D6 87 FB 54 15 3D",
                    "51 A5 1D 70 A1 C1 11 48", "00 1C C5 B7",
                     "00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F" +
                        "10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F 20 21 22 23");
    END CTRtests;

(************************************************************************)

PROCEDURE PaddingTest;

    (* Tests to see whether padding is correctly added and removed. *)

    CONST fullpad = TRUE;

    BEGIN
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "10101010101010101010101010101010");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "1010101010101010101010101010101010101010101010101010101010101010");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "03030303030303030303030303");
        DoCBChex (128, "000102030405060708090a0b0c0d0e0f",
                               "8ce82eefbea0da3c44699ed7db51b7d9",
                               "1010101010101010101010101010101003030303030303030303030303");
    END PaddingTest;

(************************************************************************)

BEGIN
    (*
    BasicTest;
    CBCtests;
    PaddingTest;
    *)
    CTRtests
END AEStest.

