(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)


MODULE DES3data;

        (********************************************************)
        (*                                                      *)
        (*         Elementary implementation of DES3            *)
        (*                                                      *)
        (*                Test of module DES                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            2 May 2018                      *)
        (*  Last edited:        3 May 2018                      *)
        (*  Status:             Complete, giving good results   *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*                                                                      *)
(*  This program implements DES3 the naive way: by doing three DES1     *)
(*  operations in sequence, precisely as specified in the standard.     *)
(*                                                                      *)
(*  Motivation: My DES implementation is well-tested and trusted, but   *)
(*  my DES3 implementation is less well trusted.  Part of the reason    *)
(*  for this is that it is hard to find test vectors for DES3 on the    *)
(*  web.  The "official" sources seem to give test vectors only for the *)
(*  special case where all three keys are the same, which is the case   *)
(*  where DES and DES3 give the same result.                            *)
(*                                                                      *)
(*  The naive DES3 implementation would be hard to get wrong, so this   *)
(*  approach serves two purposes:                                       *)
(*      (a) as a check for my more efficient implementation;            *)
(*              (result: they now agree)                                *)
(*      (b) as a way to generate test vectors for DES3.                 *)
(*                                                                      *)
(************************************************************************)


FROM SYSTEM IMPORT CARD8, ADR;

FROM DES IMPORT
    (* type *)  DEScontext,
    (* proc *)  DESopen, DESfinal, DumpKeySchedule, DESEncrypt;

FROM HexData IMPORT
    (* proc *)  HexToBytes, DumpByteArray;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  Copy;

(************************************************************************)

TYPE Block = ARRAY [0..7] OF CARD8;

(************************************************************************)

PROCEDURE SplitKey (Nkey: CARDINAL;  key: ARRAY OF CARD8;
                    VAR (*OUT*) key1, key2, key3: Block);

    (* Splits the key into three keys. *)

    BEGIN
        IF Nkey < 8 THEN
            WriteString ("ERROR: key too short");  WriteLn;
        ELSE
            Copy (ADR(key), ADR(key1), 8);
            IF Nkey >= 16 THEN
                Copy (ADR(key[8]), ADR(key2), 8);
                IF Nkey >= 24 THEN
                    Copy (ADR(key[16]), ADR(key3), 8);
                ELSE
                    key3 := key1;
                END (*IF*);
            ELSE
                key2 := key1;
                key3 := key1;
            END (*IF*);
        END (*IF*);
    END SplitKey;

(************************************************************************)

PROCEDURE DES3Encrypt (Nkey: CARDINAL;  key: ARRAY OF CARD8;
                       Ndata: CARDINAL;  data: ARRAY OF CARD8;
                       VAR (*OUT*) Nresult: CARDINAL;
                            VAR (*OUT*) result: ARRAY OF CARD8);

    (* Single encryption.  Nkey and Ndata and Nresult are   *)
    (* the number of bytes in the three arrays.             *)

    VAR outpos: CARDINAL;
        key1, key2, key3: Block;
        data2, data3: ARRAY [0..511] OF CARD8;
        ctx: DEScontext;

    BEGIN
        SplitKey (Nkey, key, key1, key2, key3);

        (* Phase 1: encryption with key1. *)

        ctx := DESopen (FALSE, FALSE, 8, key1);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data, data2, outpos);
        DESfinal (ctx, data2, outpos);
        Ndata := outpos;

        (* Phase 2: decryption with key2. *)

        ctx := DESopen (TRUE, FALSE, 8, key2);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data2, data3, outpos);
        DESfinal (ctx, data3, outpos);
        Ndata := outpos;

        (* Phase 3: encryption with key3. *)

        ctx := DESopen (FALSE, FALSE, 8, key3);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data3, result, outpos);
        DESfinal (ctx, result, outpos);
        Nresult := outpos;

    END DES3Encrypt;

(************************************************************************)

PROCEDURE DES3Decrypt (Nkey: CARDINAL;  key: ARRAY OF CARD8;
                       Ndata: CARDINAL;  data: ARRAY OF CARD8;
                       VAR (*OUT*) Nresult: CARDINAL;
                            VAR (*OUT*) result: ARRAY OF CARD8);

    (* Single decryption.  Nkey and Ndata and Nresult are   *)
    (* the number of bytes in the three arrays.             *)

    VAR outpos: CARDINAL;
        key1, key2, key3: Block;
        data2, data3: ARRAY [0..511] OF CARD8;
        ctx: DEScontext;

    BEGIN
        SplitKey (Nkey, key, key1, key2, key3);

        (* Phase 1: decryption with key3. *)

        ctx := DESopen (TRUE, FALSE, 8, key3);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data, data2, outpos);
        DESfinal (ctx, data2, outpos);
        Ndata := outpos;

        (* Phase 2: encryption with key2. *)

        ctx := DESopen (FALSE, FALSE, 8, key2);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data2, data3, outpos);
        DESfinal (ctx, data3, outpos);
        Ndata := outpos;

        (* Phase 3: decryption with key1. *)

        ctx := DESopen (TRUE, FALSE, 8, key1);
        (*DumpKeySchedule (ctx);*)
        outpos := 0;
        DESEncrypt (ctx, Ndata, data3, result, outpos);
        DESfinal (ctx, result, outpos);
        Nresult := outpos;

    END DES3Decrypt;

(************************************************************************)

PROCEDURE OneTest (key, data: ARRAY OF CHAR);

    (* Encryption and decryption using DES3-ECB.  *)

    VAR M, N, N2, Nresult: CARDINAL;
        binkey: ARRAY [0..23] OF CARD8;
        bindata, result, result2: ARRAY [0..511] OF CARD8;

    BEGIN
        WriteLn;
        M := HexToBytes (key, binkey);
        IF M < 8 THEN
            WriteString ("key is too short");  WriteLn;
            RETURN;
        END (*IF*);
        N := HexToBytes (data, bindata);
        WriteString (" key = ");  DumpByteArray (M, binkey);  WriteLn;
        WriteString ("data = ");  DumpByteArray (N, bindata);  WriteLn;
        DES3Encrypt (M, binkey, N, bindata, Nresult, result);
        WriteString ("encr = ");  DumpByteArray (Nresult, result);  WriteLn;

        (* Now do the reverse transformation. *)

        DES3Decrypt (M, binkey, Nresult, result, N2, result2);
        WriteString ("decr = ");  DumpByteArray (N2, result2);  WriteLn;

    END OneTest;

(************************************************************************)

PROCEDURE OneTestT (key, data: ARRAY OF CHAR);

    (* Same as OneTest, but the key and data are ASCII text.  *)

    VAR M, N, j, outpos: CARDINAL;
        binkey: ARRAY [0..23] OF CARD8;
        bindata, result, result2: ARRAY [0..511] OF CARD8;
        ctx: DEScontext;

    BEGIN
        WriteLn;
        M := LENGTH(key);  N := LENGTH(data);
        FOR j := 0 TO M-1 DO
            binkey[j] := ORD(key[j]);
        END (*FOR*);
        FOR j := 0 TO N-1 DO
            bindata[j] := ORD(data[j]);
        END (*FOR*);
        WriteString (" key = ");  DumpByteArray (M, binkey);  WriteLn;
        WriteString ("data = ");  DumpByteArray (N, bindata);  WriteLn;
        ctx := DESopen (FALSE, FALSE, M, binkey);

        (*DumpKeySchedule (ctx);*)

        outpos := 0;
        DESEncrypt (ctx, N, bindata, result, outpos);
        DESfinal (ctx, result, outpos);
        N := outpos;
        WriteString ("encr = ");  DumpByteArray (outpos, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := DESopen (TRUE, FALSE, M, binkey);
        outpos := 0;
        DESEncrypt (ctx, N, result, result2, outpos);
        DESfinal (ctx, result2, outpos);
        WriteString ("decr = ");  DumpByteArray (outpos, result2);  WriteLn;

    END OneTestT;

(************************************************************************)

PROCEDURE DEStests;

    (* Run some test vectors.  I am taking these from                   *)
    (* https://www.cosic.esat.kuleuven.be/nessie/testvectors/bc/des/Des-64-64.test-vectors  *)
    (* That's a big test suite, so I've only taken a sample from it.    *)

    BEGIN
        (* Some from set 1. *)

        OneTest ("8000000000000000", "0000000000000000");
        OneTest ("4000000000000000", "0000000000000000");
        OneTest ("2000000000000000", "0000000000000000");
        OneTest ("1000000000000000", "0000000000000000");
        OneTest ("0800000000000000", "0000000000000000");
        OneTest ("0400000000000000", "0000000000000000");
        OneTest ("0200000000000000", "0000000000000000");
        OneTest ("0100000000000000", "0000000000000000");

        (* A sampling from set 3, which exercises more keys. *)

        OneTest ("0000000000000000", "0000000000000000");
                        (* result should be 8c a6 4d e9 c1 b1 23 a7 *)
        OneTest ("0101010101010101", "0101010101010101");
                        (* result should be 99 4d 4d c1 57 b9 6c 52 *)
        OneTest ("3A3A3A3A3A3A3A3A", "3A3A3A3A3A3A3A3A");   (* vector #58 *)
                        (* result should be b0 4a 2a ac 92 5a bb 0b *)

        (* Set 4. *)

        OneTest ("0001020304050607", "0011223344556677");
                        (* result should be 3e f0 a8 91 cf 8e d9 90 *)
        OneTest ("2BD6459F82C5B300", "EA024714AD5C4D84");
                        (* result should be 12 6e fe 8e d3 12 19 0a *)

    END DEStests;

(************************************************************************)

PROCEDURE DES3tests;

    (* Test vectors for DES3 are hard to find, but I'll use what I can. *)

    (* Results so far: in only one of the tests below there is a        *)
    (* discrepancy between my result and a result published on the web. *)
    (* Since we are passing all other tests, I feel safe in concluding  *)
    (* that that one case came from a faulty source.                    *)

    BEGIN
        WriteLn;  WriteString ("--------  DES3 TESTS  --------");  WriteLn;

        (* DES3 with three equal keys should give same result as DES. *)

        OneTest ("8000000000000000", "0000000000000000");
                        (* result should be 95A8D72813DAA94D, and it is. *)
        OneTest ("75 28 78 39 74 93 CB 70", "11 22 33 44 55 66 77 88");
                        (* result should be B5 21 9E E8 1A A7 49 9D, and it is. *)

        (* DES3 with two equal keys should also give same result as DES. *)

        OneTest ("80000000000000008000000000000000", "0000000000000000");
                        (* and it does *)

        (* Now a test with two unequal keys. *)

        OneTest ("80000000000000000000000000000000", "0000000000000000");
                        (* inverts OK, result is fa fd 50 84 37 4f ce 34 *)
        OneTest ("00000000000000008000000000000000", "0000000000000000");
                        (* inverts OK, result is c2 a4 dd 96 15 14 53 c2 *)
        OneTest ("75 28 78 39 74 93 CB 70 34 E5 46 7B 95 94 40 6D", "11 22 33 44 55 66 77 88");
                        (* inverts OK, result is c0 f4 98 05 7d 93 5d 6a *)
                        (* result should be (?) 74 3F 14 0C F3 72 B3 3C *)

        (* Three unequal keys. *)

        OneTest ("400000000000000000000000000000008000000000000000", "0000000000000000");
                        (* result is 07 5b bf ff f4 9a 51 34 *)

        (* This next one is taken from                                  *)
        (*        https://asecuritysite.com/encryption/threedes         *)

        OneTestT ("1234567890123456ABCDEFGH", "The quic");
                        (* inverts OK, result is 13 d4 d3 54 94 93 d2 87 *)
                        (* result should be 13D4D3549493D287 *)

    END DES3tests;

(************************************************************************)

PROCEDURE PaddingTests;

    (* Tests with a non-integral number of blocks.  *)

    BEGIN
        OneTest ("0000000000000000", "0000000000000000");
                        (* result should be 8c a6 4d e9 c1 b1 23 a7 *)
        OneTestT ("1234567890123456ABCDEFGH", "The quick brown fox jumped over the lazy dog");
                        (* inverts OK, result agrees with the one below *)
                        (* result should be (?) 13D4D3549493D2870F93C3E0812A06DE467E1F9C0BFB16C070EDE5CABBD3CA62F217A7AE8D47F2C7198FF39162ECEC94 *)
    END PaddingTests;

(************************************************************************)

PROCEDURE TroubleCase;

    BEGIN
        (* No current troublesome cases. *)
    END TroubleCase;

(************************************************************************)

BEGIN
    (*
    DEStests;
    *)
    DES3tests;
    PaddingTests;
    TroubleCase;
END DES3data.

