(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)


MODULE DEStest;

        (********************************************************)
        (*                                                      *)
        (*           DES: Data encryption standard              *)
        (*                                                      *)
        (*                Test of module DES                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            24 March 2018                   *)
        (*  Last edited:        24 January 2021                 *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT DES;

FROM SYSTEM IMPORT CARD8, ADR;

FROM DES IMPORT
    (* type *)  DEScontext,
    (* proc *)  DESopen, DES_CBCopen, DESfinal, DumpKeySchedule, DESEncrypt;

FROM HexData IMPORT
    (* proc *)  HexToBytes, DumpByteArray;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  Copy;

(************************************************************************)

VAR fullpadding: BOOLEAN;

(************************************************************************)

PROCEDURE OneCBCTest (fullpadding: BOOLEAN;  key, data, IV: ARRAY OF CHAR);

    (* Encryption and decryption using DES3-CBC.    *)
    (* All parameters are ASCII.                    *)

    VAR M, N, P, j, outpos: CARDINAL;
        binkey: ARRAY [0..23] OF CARD8;
        binIV: ARRAY [0..7] OF CARD8;
        bindata, result, result2: ARRAY [0..2047] OF CARD8;
        ctx: DEScontext;

    BEGIN
        WriteString ("------------------------");  WriteLn;
        M := LENGTH(key);  N := LENGTH(data);  P := LENGTH(IV);
        FOR j := 0 TO M-1 DO
            binkey[j] := ORD(key[j]);
        END (*FOR*);
        FOR j := 0 TO N-1 DO
            bindata[j] := ORD(data[j]);
        END (*FOR*);
        FOR j := 0 TO P-1 DO
            binIV[j] := ORD(IV[j]);
        END (*FOR*);
        WriteString (" key = ");  WriteString (key);  WriteLn;
        WriteString ("data = ");  WriteString (data);  WriteLn;
        WriteString ("     = ");  DumpByteArray (N, bindata);  WriteLn;
        WriteString (" IV  = ");  WriteString (IV);  WriteLn;
        ctx := DES_CBCopen (FALSE, fullpadding, M, binkey, binIV);
        outpos := 0;
        DESEncrypt (ctx, N, bindata, result, outpos);
        DESfinal (ctx, result, outpos);
        N := outpos;
        WriteString ("encr = ");  DumpByteArray (outpos, result);  WriteLn;

        (* Now do the reverse transformation. *)

        outpos := 0;
        ctx := DES_CBCopen (TRUE, fullpadding, M, binkey, binIV);
        DESEncrypt (ctx, N, result, result2, outpos);
        DESfinal (ctx, result2, outpos);
        WriteString ("decr = ");  DumpByteArray (outpos, result2);  WriteLn;

    END OneCBCTest;

(************************************************************************)

PROCEDURE OneTest (key, data: ARRAY OF CHAR);

    (* Encryption and decryption using DES3-ECB.  *)

    VAR M, N, outpos: CARDINAL;
        binkey: ARRAY [0..23] OF CARD8;
        bindata, result, result2: ARRAY [0..511] OF CARD8;
        ctx: DEScontext;

    BEGIN
        WriteLn;
        M := HexToBytes (key, binkey);
        IF M < 8 THEN
            WriteString ("key is too short");  WriteLn;
            RETURN;
        END (*IF*);
        N := HexToBytes (data, bindata);
        WriteString (" key = ");  DumpByteArray (M, binkey);  WriteLn;
        WriteString ("data = ");  DumpByteArray (N, bindata);  WriteLn;
        ctx := DESopen (FALSE, fullpadding, M, binkey);

        (* Test of some internal detail. *)

        DumpKeySchedule (ctx);

        outpos := 0;
        DESEncrypt (ctx, N, bindata, result, outpos);
        DESfinal (ctx, result, outpos);
        N := outpos;
        WriteString ("encr = ");  DumpByteArray (outpos, result);  WriteLn;

        (* Now do the reverse transformation. *)

        outpos := 0;
        ctx := DESopen (TRUE, fullpadding, M, binkey);
        DESEncrypt (ctx, N, result, result2, outpos);
        DESfinal (ctx, result2, outpos);
        WriteString ("decr = ");  DumpByteArray (outpos, result2);  WriteLn;

    END OneTest;

(************************************************************************)

PROCEDURE OneTestT (key, data: ARRAY OF CHAR);

    (* Same as OneTest, but the key and data are ASCII text.  *)

    VAR M, N, j, outpos: CARDINAL;
        binkey: ARRAY [0..23] OF CARD8;
        bindata, result, result2: ARRAY [0..511] OF CARD8;
        ctx: DEScontext;

    BEGIN
        WriteLn;
        M := LENGTH(key);  N := LENGTH(data);
        FOR j := 0 TO M-1 DO
            binkey[j] := ORD(key[j]);
        END (*FOR*);
        FOR j := 0 TO N-1 DO
            bindata[j] := ORD(data[j]);
        END (*FOR*);
        WriteString (" key = ");  DumpByteArray (M, binkey);  WriteLn;
        WriteString ("data = ");  DumpByteArray (N, bindata);  WriteLn;
        ctx := DESopen (FALSE, fullpadding, M, binkey);

        (* Test of some internal detail. *)

        DumpKeySchedule (ctx);

        outpos := 0;
        DESEncrypt (ctx, N, bindata, result, outpos);
        DESfinal (ctx, result, outpos);
        N := outpos;
        WriteString ("encr = ");  DumpByteArray (outpos, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := DESopen (TRUE, fullpadding, M, binkey);
        outpos := 0;
        DESEncrypt (ctx, N, result, result2, outpos);
        DESfinal (ctx, result2, outpos);
        WriteString ("decr = ");  DumpByteArray (outpos, result2);  WriteLn;

    END OneTestT;

(************************************************************************)

PROCEDURE DEStests;

    (* Run some test vectors.  I am taking these from                   *)
    (* https://www.cosic.esat.kuleuven.be/nessie/testvectors/bc/des/Des-64-64.test-vectors  *)
    (* That's a big test suite, so I've only taken a sample from it.    *)

    BEGIN
        (* Some from set 1. *)

        OneTest ("8000000000000000", "0000000000000000");
        OneTest ("4000000000000000", "0000000000000000");
        OneTest ("2000000000000000", "0000000000000000");
        OneTest ("1000000000000000", "0000000000000000");
        OneTest ("0800000000000000", "0000000000000000");
        OneTest ("0400000000000000", "0000000000000000");
        OneTest ("0200000000000000", "0000000000000000");
        OneTest ("0100000000000000", "0000000000000000");

        (* A sampling from set 3, which exercises more keys. *)

        OneTest ("0000000000000000", "0000000000000000");
                        (* result should be 8c a6 4d e9 c1 b1 23 a7 *)
        OneTest ("0101010101010101", "0101010101010101");
                        (* result should be 99 4d 4d c1 57 b9 6c 52 *)
        OneTest ("3A3A3A3A3A3A3A3A", "3A3A3A3A3A3A3A3A");   (* vector #58 *)
                        (* result should be B04A2AAC925ABB0B *)

        (* Set 4. *)

        OneTest ("0001020304050607", "0011223344556677");
                        (* result should be 3e f0 a8 91 cf 8e d9 90 *)
        OneTest ("2BD6459F82C5B300", "EA024714AD5C4D84");
                        (* result should be 12 6e fe 8e d3 12 19 0a *)

    END DEStests;

(************************************************************************)

PROCEDURE DES3tests;

    (* Test vectors for DES3 are hard to find, but I'll use what I can. *)

    (* Results so far: there is one remaining case where my encryption  *)
    (* disagrees with the result claimed by a web source.  After        *)
    (* carefully doing the job a different way (see program DES3data),  *)
    (* I think I can safely say that that web source is probably at     *)
    (* fault.  The basic difficulty is that officially published test   *)
    (* vectors focus on the case of three equal keys, so my test cases  *)
    (* come from searching the web for data from sources of uncertain   *)
    (* reliability.                                                     *)

    BEGIN
        WriteLn;  WriteString ("--------  DES3 TESTS  --------");  WriteLn;

        (* DES3 with three equal keys should give same result as DES. *)

        OneTest ("8000000000000000", "0000000000000000");
                        (* result should be 95A8D72813DAA94D, and it is. *)
        OneTest ("75 28 78 39 74 93 CB 70", "11 22 33 44 55 66 77 88");
                        (* result should be B5 21 9E E8 1A A7 49 9D, and it is. *)

        (* DES3 with two equal keys should also give same result as DES. *)

        OneTest ("80000000000000008000000000000000", "0000000000000000");
                        (* and it does *)

        (* Now a test with two unequal keys. *)

        OneTest ("80000000000000000000000000000000", "0000000000000000");
                        (* inverts OK, result is fa fd 50 84 37 4f ce 34 *)
        OneTest ("00000000000000008000000000000000", "0000000000000000");
                        (* inverts OK, result is c2 a4 dd 96 15 14 53 c2 *)

        OneTest ("75 28 78 39 74 93 CB 70 34 E5 46 7B 95 94 40 6D", "11 22 33 44 55 66 77 88");
                        (* inverts OK, result is c0 f4 98 05 7d 93 5d 6a *)
                        (* result should be (?) 74 3F 14 0C F3 72 B3 3C *)

        (* Three unequal keys. *)

        OneTest ("400000000000000000000000000000008000000000000000", "0000000000000000");
                        (* result is 07 5b bf ff f4 9a 51 34 *)

        (* This next one is taken from                                  *)
        (*       https://asecuritysite.com/encryption/threedes          *)

        OneTestT ("1234567890123456ABCDEFGH", "The quic");
                        (* inverts OK, result is 13 d4 d3 54 94 93 d2 87 *)
                        (* result should be 13D4D3549493D287 *)

    END DES3tests;

(************************************************************************)

PROCEDURE PaddingTests;

    (* Tests with a non-integral number of blocks.  *)

    BEGIN
        fullpadding := FALSE;
        OneTest ("0000000000000000", "0000000000000000");
                        (* result should be 8c a6 4d e9 c1 b1 23 a7 *)
        OneTestT ("1234567890123456ABCDEFGH", "The quick brown fox jumped over the lazy dog");
                        (* result no agrees with my web source, bug fixed *)
                        (* result should be 13D4D3549493D2870F93C3E0812A06DE467E1F9C0BFB16C070EDE5CABBD3CA62F217A7AE8D47F2C7198FF39162ECEC94 *)
    END PaddingTests;

(************************************************************************)

PROCEDURE CBCtests;

    (* Test vectors for DES3-CBC are even harder to find.  The source   *)
    (* I am using is https://asecuritysite.com/encryption/              *)
    (* It appears that it is using fullpadding = TRUE, so I'll have to  *)
    (* do the same to be able to compare.                               *)

    BEGIN
        WriteLn;  WriteString ("--------  DES3-CBC TESTS  --------");  WriteLn;

        OneCBCTest (TRUE, "1234567890123456ABCDEFGH",
            "The quick brown fox jumped over the lazy dog....", "12345678");
        OneCBCTest (TRUE, "1234567890123456ABCDEFGH",
            "The quick brown fox jumped over the lazy dog", "12345678");
        OneCBCTest (FALSE, "1234567890123456ABCDEFGH",
            "The quick brown fox jumped over the lazy dog....", "12345678");
        OneCBCTest (FALSE, "1234567890123456ABCDEFGH",
            "The quick brown fox jumped over the lazy dog", "12345678");
        OneCBCTest (TRUE, "1234567890123456ABCDEFGH",
            "The quick brown fox jumped over the lazy dog", "23456789");

        OneCBCTest (TRUE, "1234567890123456ABCDEFGH",
            "Humpty Dumpty sat on the wall", "23456789");

        (* All tests passed so far. *)

        (* I get disagreement when I use shorter keys, but I can't      *)
        (* tell whether my web source is buggy for that case.           *)

    END CBCtests;

(************************************************************************)

BEGIN
    fullpadding := TRUE;
    (*
    DEStests;
    (**)
    DES3tests;
    (**)
    PaddingTests;
    *)
    CBCtests;

END DEStest.

