(**************************************************************************)
(*                                                                        *)
(*  Test of HMAC keyed encryption                                         *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE HMAC_SHA1Test;

        (********************************************************)
        (*                                                      *)
        (*              TEST OF HMAC-SHA1 ENCRYPTION            *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            30 September 2017               *)
        (*  Last edited:        6 October 2017                  *)
        (*  Status:             Passes all tests                *)
        (*                                                      *)
        (*  The test data are taken from RFC 2202.              *)
        (*                                                      *)
        (********************************************************)


IMPORT STextIO;

FROM SYSTEM IMPORT LOC, CARD8, CARD32;

FROM SHA1 IMPORT
    (* type *)  SHA1_DigestType;

FROM BEDigests IMPORT
    (* proc *)  BEDigestToString;

FROM HMAC IMPORT
    (* proc *)  HMAC_SHA1;

(************************************************************************)

PROCEDURE OneEncryptionSet (text: ARRAY OF LOC;  textlength: CARDINAL;
                            key: ARRAY OF LOC;  keylength: CARDINAL);

    VAR digest1: SHA1_DigestType;
        result: ARRAY [0..63] OF CHAR;

    BEGIN
        (* HMAC_SHA1 *)

        HMAC_SHA1 (key, keylength, text, textlength, digest1);
        BEDigestToString (5, digest1, result);
        STextIO.WriteString ("SHA1:   ");
        STextIO.WriteString (result);
        STextIO.WriteLn;

    END OneEncryptionSet;

(************************************************************************)

PROCEDURE Hexval (ch: CHAR): CARDINAL;

    (* Convert hex char to binary.  We trust that ch only takes values  *)
    (* in '0'..'9' or 'a'..'f'.                                         *)

    BEGIN
        IF ch > '9' THEN
            RETURN ORD(ch) - ORD('a') + 10;
        ELSE
            RETURN ORD(ch) - ORD('0');
        END (*IF*);
    END Hexval;

(************************************************************************)

PROCEDURE Decode (hexdata: ARRAY OF CHAR;  VAR (*OUT*) bytes: ARRAY OF CARD8): CARDINAL;

    (* Converts a hexadecimal string (2 chars/value) to a string of     *)
    (* bytes.  The function result is the length of the result.         *)

    VAR k, N: CARDINAL;

    BEGIN
        N := LENGTH (hexdata);
        IF N > 1 THEN
            FOR k := 0 TO N-2 BY 2 DO
                bytes[k DIV 2] := 16*Hexval(hexdata[k]) + Hexval(hexdata[k+1]);
            END (*FOR*);
        END (*IF*);
        RETURN N DIV 2;
    END Decode;

(************************************************************************)

PROCEDURE DoTestLL (testnum: CARDINAL;  key, data: ARRAY OF CHAR);

    (* The key and data are both ASCII.          *)

    VAR datalength, keylength: CARDINAL;

    BEGIN
        STextIO.WriteString ("Test case ");
        STextIO.WriteChar (CHR(ORD('0') + testnum));
        STextIO.WriteLn;
        keylength := LENGTH (key);
        datalength := LENGTH (data);
        OneEncryptionSet (data, datalength, key, keylength);
    END DoTestLL;

(************************************************************************)

PROCEDURE DoTestCL (testnum: CARDINAL;  key_enc, data: ARRAY OF CHAR);

    (* The key is in the form of hexadecimal strings, with two hex      *)
    (* characters representing each octet.  The data is ASCII.          *)

    VAR datalength, keylength: CARDINAL;
        key: ARRAY [0..255] OF CARD8;

    BEGIN
        STextIO.WriteString ("Test case ");
        STextIO.WriteChar (CHR(ORD('0') + testnum));
        STextIO.WriteLn;
        keylength := Decode (key_enc, key);
        datalength := LENGTH (data);
        OneEncryptionSet (data, datalength, key, keylength);
    END DoTestCL;

(************************************************************************)

PROCEDURE DoTestCC (testnum: CARDINAL;  key_enc, data_enc: ARRAY OF CHAR);

    (* The test data we are given to work with are in the form of       *)
    (* hexadecimal strings, with two hex characters representing each   *)
    (* octet.  We must of course decode that information.               *)

    VAR datalength, keylength: CARDINAL;
        data, key: ARRAY [0..255] OF CARD8;

    BEGIN
        STextIO.WriteString ("Test case ");
        STextIO.WriteChar (CHR(ORD('0') + testnum));
        STextIO.WriteLn;
        keylength := Decode (key_enc, key);
        datalength := Decode (data_enc, data);
        OneEncryptionSet (data, datalength, key, keylength);
    END DoTestCC;

(************************************************************************)

PROCEDURE RunTheTests;

    (* Runs the 7 test cases given in RFC 2202. *)

    CONST Nul = CHR(0);

    VAR j: CARDINAL;
        arr1: ARRAY [0..131] OF CHAR;
        arr2: ARRAY [0..50] OF CHAR;

    BEGIN
        STextIO.WriteString ("TEST OF HMAC-SHA1");  STextIO.WriteLn;

        (* Test case 1 *)

        DoTestCL (1, "0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b", "Hi There");

        (* Test case 2 *)

        DoTestLL (2, "Jefe", "what do ya want for nothing?");

        (* Test case 3 *)

        FOR j := 0 TO 49 DO
            arr2[j] := CHR(0DDH);
        END (*FOR*);
        arr2[50] := Nul;
        DoTestCL (3, "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", arr2);

        (* Test case 4 *)

        FOR j := 0 TO 49 DO
            arr2[j] := CHR(0CDH);
        END (*FOR*);
        arr2[50] := Nul;
        DoTestCL (4, "0102030405060708090a0b0c0d0e0f10111213141516171819", arr2);

        (* Test case 5 *)

        FOR j := 0 TO 19 DO
            arr1[j] := CHR(0CH);
        END (*FOR*);
        arr1[20] := Nul;
        DoTestLL (5, arr1, "Test With Truncation");

        (* Test case 6 *)

        FOR j := 0 TO 79 DO
            arr1[j] := CHR(0AAH);
        END (*FOR*);
        arr1[80] := Nul;
        DoTestLL (6, arr1, "Test Using Larger Than Block-Size Key - Hash Key First");

        (* Test case 7 *)

        (* Use same arr1 as in case 6. *)
        DoTestLL (7, arr1, "Test Using Larger Than Block-Size Key"
                            + " and Larger Than One Block-Size Data");

    END RunTheTests;

(************************************************************************)

BEGIN
    RunTheTests;
END HMAC_SHA1Test.
