MODULE RC4Test;

        (********************************************************)
        (*                                                      *)
        (*                TEST OF RC4 ENCRYPTION                *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            17 April 2018                   *)
        (*  Last edited:        18 April 2018                   *)
        (*  Status:             Passes all tests                *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*                                                                      *)
(* This test program is based on test data in                           *)
(*              https://en.wikipedia.org/wiki/RC4                       *)
(* and also tests in RFC 6229                                           *)
(*                                                                      *)
(************************************************************************)

FROM SYSTEM IMPORT CARD8;

IMPORT Strings;

FROM RC4 IMPORT
    (* type *)  RC4state,
    (* proc *)  RC4Init, Encrypt, RC4Close;

FROM HexData IMPORT
    (* proc *)  HexToBytes, DumpByteArray;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(********************************************************************************)
(*                                 OUTPUT                                       *)
(********************************************************************************)

PROCEDURE WriteCardN (A, N, threshold: CARDINAL);

    (* Writes A as an N-digit decimal number, right justified.  The threshold   *)
    (* parameter says how big A has to be to need all N digits.                 *)

    BEGIN
        IF N <= 1 THEN
            WriteChar (CHR(ORD('0') + A));
        ELSIF A >= threshold THEN
            WriteCardN (A DIV 10, N-1, threshold DIV 10);
            WriteCardN (A MOD 10, 1, 0);
        ELSE
            WriteChar (' ');
            WriteCardN (A, N-1, threshold DIV 10);
        END (*IF*);
    END WriteCardN;

(********************************************************************************)

PROCEDURE WriteCard4 (A: CARDINAL);

    (* Writes A as a 4-digit decimal number, right justified.  *)

    BEGIN
        WriteCardN (A, 4, 1000);
    END WriteCard4;

(********************************************************************************)
(*                             THE ENCRYPTION TEST                              *)
(********************************************************************************)

PROCEDURE ATest (key, text: ARRAY OF CHAR);

    (* In this version the key and text are string literals. *)

    VAR length: CARDINAL;  state: RC4state;
        output: ARRAY [0..1023] OF CARD8;
        final: ARRAY [0..1023] OF CHAR;

    BEGIN
        state := RC4Init (0, key, Strings.Length(key));
        WriteString ("key = ");
        WriteString (key);  WriteString (', ');
        WriteString ("text = ");
        WriteString (text);
        WriteLn;
        length := Strings.Length(text);
        Encrypt (state, length, text, output);
        WriteString ("encrypted = ");
        DumpByteArray (length, output);
        WriteLn;
        RC4Close (state);

        (* Now do the reverse operation. *)

        state := RC4Init (0, key, Strings.Length(key));
        Encrypt (state, length, output, final);
        final[length] := CHR(0);
        WriteString ("decrypted = ");
        WriteString (final);
        WriteLn;
        RC4Close (state);

    END ATest;

(************************************************************************)

PROCEDURE BTest (key, text: ARRAY OF CHAR);

    (* In this version the key and text are strings to be interpreted   *)
    (* as hexadecimal numbers.  In addition, we run the test with       *)
    (* several different initial offsets.                               *)

    CONST Noff = 9;
    TYPE Offsets = ARRAY [0..Noff-1] OF CARDINAL;
    CONST offset = Offsets {0, 256, 512, 768, 1024, 1536, 2048, 3072, 4096};

    VAR case, Nk, Nt: CARDINAL;  state: RC4state;
        binkey, bintext: ARRAY [0..1023] OF CARD8;
        output: ARRAY [0..1023] OF CARD8;

    BEGIN
        WriteLn;
        WriteString ("key = ");
        WriteString (key);
        (*  WriteString (', ');  WriteString ("text = ");  WriteString (text); *)
        WriteLn;
        Nk := HexToBytes (key, binkey);
        Nt := HexToBytes (text, bintext);
        FOR case := 0 TO Noff-1 DO
            state := RC4Init (offset[case], binkey, Nk);
            Encrypt (state, Nt, bintext, output);
            WriteCard4 (offset[case]);  WriteString (":  ");
            DumpByteArray (Nt, output);
            WriteLn;
            RC4Close (state);
        END (*FOR*);

        (* Don't bother decrypting, we trust that now. *)

    END BTest;

(************************************************************************)

PROCEDURE DoTests;

    BEGIN
        (* From the Wikipedia article.  *)

        ATest ("Key", "Plaintext");
        ATest ("Wiki", "pedia");
        ATest ("Secret", "Attack at dawn");

        (* From RFC 6229. *)

        BTest ("0102030405", "00000000 00000000 00000000 00000000");
        BTest ("01020304050607", "00000000 00000000 00000000 00000000");
        BTest ("0102030405060708", "00000000 00000000 00000000 00000000");
        BTest ("0102030405060708090a", "00000000 00000000 00000000 00000000");
        BTest ("0102030405060708090a0b0c0d0e0f10", "00000000 00000000 00000000 00000000");
        BTest ("0102030405060708090a0b0c0d0e0f101112131415161718", "00000000 00000000 00000000 00000000");
        BTest ("0102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f20", "00000000 00000000 00000000 00000000");

        BTest ("833222772a", "00000000 00000000 00000000 00000000");
        BTest ("1910833222772a", "00000000 00000000 00000000 00000000");
        BTest ("641910833222772a", "00000000 00000000 00000000 00000000");
        BTest ("8b37641910833222772a", "00000000 00000000 00000000 00000000");
        BTest ("ebb46227c6cc8b37641910833222772a", "00000000 00000000 00000000 00000000");
        BTest ("c109163908ebe51debb46227c6cc8b37641910833222772a", "00000000 00000000 00000000 00000000");
        BTest ("1ada31d5cf688221c109163908ebe51debb46227c6cc8b37641910833222772a", "00000000 00000000 00000000 00000000");

    END DoTests;

(************************************************************************)
(*                          MAIN TEST PROGRAM                           *)
(************************************************************************)

BEGIN
    DoTests;
END RC4Test.

