(**************************************************************************)
(*                                                                        *)
(*  A TLS test client                                                     *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE TLSClient;

        (********************************************************)
        (*                                                      *)
        (*                   TLS test client                    *)
        (*         Used to exercise my TLS server code          *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            8 August 2018                   *)
        (*  Last edited:        12 August 2018                  *)
        (*  Status:             Just started                    *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM TransLog IMPORT
    (* type *)  LogContext, TransactionLogID,
    (* proc *)  OpenLogContext, CloseLogContext, StartTransactionLogging, CreateLogID, DiscardLogID,
                LogTransaction, LogTransactionL;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Inet2Misc IMPORT
    (* proc *)  Swap2, StringToIP;

FROM Sockets IMPORT
    (* const*)  NotASocket,
    (* type *)  Socket, SockAddr, AddressFamily, SocketType,
    (* proc *)  sock_init, sock_errno, socket, connect, soclose;

FROM Internet IMPORT
    (* const*)  Zero8, INADDR_ANY;

FROM NetStream IMPORT
    (* type *)  NStream,
    (* proc *)  OpenNetStream;

(************************************************************************)

VAR
    LContext: LogContext;

(************************************************************************)
(*                           THE MAIN CLIENT CODE                       *)
(************************************************************************)

(*
PROCEDURE WriteError (LogID: TransactionLogID);

    VAR LogLine: ARRAY [0..255] OF CHAR;

    BEGIN
        Strings.Assign ("Socket error ", LogLine);
        AppendCard (sock_errno(), LogLine);
        LogTransaction (LogID, LogLine);
    END WriteError;
*)

(************************************************************************)

PROCEDURE ConnectToHost (IPaddress: CARDINAL;  ServerPort: CARDINAL;
                      VAR (*INOUT*) FailureReason: ARRAY OF CHAR): Socket;

    (* Tries to open a connection to the specified host.  Returns the   *)
    (* value NotASocket if we don't succeed; in this case, the string   *)
    (* FailureReason is given a value.                                  *)
    (* IPaddress is in network byte order.                              *)

    VAR s: Socket;  peer: SockAddr;

    BEGIN
        IF IPaddress <> 0 THEN
            s := socket (AF_INET, SOCK_STREAM, AF_UNSPEC);

            IF s = NotASocket THEN
                Strings.Assign ("Can't allocate socket", FailureReason);
            ELSE

                (* Bind to an address at our end if we're using a       *)
                (* specific address.                                    *)

                (*
                IF BindAddr <> 0 THEN
                    WITH peer DO
                        family := AF_INET;
                        WITH in_addr DO
                            port := 0;
                            addr := BindAddr;
                            zero := Zero8;
                        END (*WITH*);
                    END (*WITH*);

                    IF bind (s, peer, SIZE(peer)) THEN
                        WriteError (LogID);
                        LogTransactionL (LogID, "Cannot bind to local interface");
                    END (*IF*);
                END (*IF*);
                *)

                (* Socket open, connect to the client. *)

                WITH peer DO
                    family := AF_INET;
                    WITH in_addr DO
                        port := Swap2(ServerPort);
                        addr := IPaddress;
                        zero := Zero8;
                    END (*WITH*);
                END (*WITH*);

                IF connect (s, peer, SIZE(peer)) THEN

                    Strings.Assign ("Failed to connect", FailureReason);
                    soclose(s);
                    s := NotASocket;

                END (*IF*);
            END (*IF*);

        ELSE

            Strings.Assign ("500 Unknown host", FailureReason);
            s := NotASocket;

        END (*IF*);

        RETURN s;

    END ConnectToHost;

(********************************************************************************)

PROCEDURE RunTheClient;

    (*  OPERATING AS A CLIENT                                                       *)
    (*     1. Create a new socket.                                                  *)
    (*     2. (Optional) Call "bind" to bind the socket with a local address.       *)
    (*        This is normally not needed.                                          *)
    (*     3. Call "connect" to connect to the server.                              *)
    (*     4. If you want a secure connection, do the TLS negotiation.              *)
    (*     5. For a non-secure connection, use procedures "send" and "recv" to      *)
    (*        transfer data.  In the secure case, use my module NetStream to        *)
    (*        transfer using encryption; it still uses send and recv, but some      *)
    (*        transformation of the data happens along the way.                     *)
    (*     6. Use "soclose" on your original socket to clean up at the end.         *)
    (*  If using NetStream, steps 4, 5, and 6 are all handled inside module         *)
    (*  NetStream, so all we have to worry about here are the socket creation and   *)
    (*  the "connect" operation.                                                    *)

    CONST useTLS = TRUE;
        ServerName = "127.0.0.1";
        ServerPort = 443;

    VAR
        LogID: TransactionLogID;
        ServerAddress: CARDINAL;
        s: Socket;
        stream: NStream;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        (* Initial setup and opening log messages. *)

        LogID := CreateLogID (LContext, "       ");
        IF sock_init() <> 0 THEN
            LogTransactionL (LogID, "No network.");
            DiscardLogID (LogID);
            RETURN;
        END (*IF*);

        ServerAddress := StringToIP (ServerName);
        s := ConnectToHost (ServerAddress, ServerPort, message);
        IF s = NotASocket THEN
            LogTransactionL (LogID, message);
            RETURN;
        END (*IF*);
        LogTransactionL (LogID, "Have connected to server at socket level");

        stream := OpenNetStream (s, FALSE, LogID, TRUE);

    END RunTheClient;

(********************************************************************************)
(*                                PROGRAM START                                 *)
(********************************************************************************)

BEGIN
    LContext := OpenLogContext();
    StartTransactionLogging (LContext, "TLSClient.log", 2);
    RunTheClient;
    CloseLogContext (LContext);
END TLSClient.

