MODULE testDH;

        (********************************************************)
        (*                                                      *)
        (*       Simulation of Diffie-Helman key exchange       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        19 January 2021                 *)
        (*  Status:             Work in progress                *)
        (*                                                      *)
        (********************************************************)


IMPORT BigNum;

FROM RSA IMPORT
    (* proc *)  Encrypt, Decrypt;

FROM RSAKeys IMPORT
    (* type *)  RSAKeyType, DiscardKey;

FROM RSAGenerate IMPORT
    (* proc *)  RSA_Generate;

FROM BigNum IMPORT
    (* type *)  BN;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)

VAR
    generator, Oakley2, Oakley14: BN;

(************************************************************************)
(*                             TEST FUNCTIONS                           *)
(************************************************************************)

PROCEDURE Oakley;

    (* This simply sets the prime and generator for Oakley Group 2 and  *)
    (* Oakley Group 14.  Values are from RFC 2409.                      *)

    VAR check: ARRAY [0..512] OF CHAR;

    BEGIN
        generator := BigNum.MakeBignum (2);

        Oakley2 := BigNum.HexToBN (
                   "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
                 + "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
                 + "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
                 + "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
                 + "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE65381"
                 + "FFFFFFFF FFFFFFFF");

        WriteString ("The Oakley 2 prime is");  WriteLn;
        BigNum.ToHex (Oakley2, check, 257);
        WriteString (check);  WriteLn;

        Oakley14 := BigNum.HexToBN (
                      "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
                    + "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
                    + "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
                    + "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
                    + "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D"
                    + "C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F"
                    + "83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D"
                    + "670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B"
                    + "E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9"
                    + "DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510"
                    + "15728E5A 8AACAA68 FFFFFFFF FFFFFFFF");

        WriteString ("The Oakley 14 prime is");  WriteLn;
        BigNum.ToHex (Oakley14, check, 513);
        WriteString (check);  WriteLn;

    END Oakley;

(************************************************************************)

PROCEDURE Simulate (variant: CARDINAL);

    (* Simulation of the key exchange algorithm. *)

    CONST privatesize = 14;

    VAR kclient, kserver: RSAKeyType;
        prime, Ksharedc, Kshareds: BN;

    BEGIN
        IF variant = 14 THEN prime := Oakley14;
        ELSE prime := Oakley2;
        END (*IF*);

        (* Initially each side sends name-lists of supported        *)
        (* algorithms.  I'm not simulating that step.               *)

        (* Key exchange begins by the client generating an ephemeral    *)
        (* keypair (private and associated public key) and sending the  *)
        (* server its public key in a SSH_MSG_KEX_ECDH_INIT message.    *)
        (* This keypair is EPHEMERAL: it will only be used during the   *)
        (* key exchange and disposed of afterwards.                     *)

        (* Client chooses random number a and sends A = g^a mod p,      *)
        (* where g and p are the Oakley parameters.  A becomes the      *)
        (* client's public key.                                         *)

        kclient.private := BigNum.Rand(privatesize);
        kclient.public := BigNum.ModularPower (generator, kclient.private, prime);
        WriteString ("Client private key is");  WriteLn;
        BigNum.WriteBignum (kclient.private);  WriteLn;
        WriteString ("Client public key is");  WriteLn;
        BigNum.WriteBignum (kclient.public);  WriteLn;

        (* The server listens for the SSH_MSG_KEX_ECDH_INIT message,    *)
        (* and upon receipt, generates its own ephemeral keypair. With  *)
        (* the client's public key and its own keypair, the server can  *)
        (*  generate the shared secret K.                               *)

        (* Server does a similar calculation.  *)

        kserver.private := BigNum.Rand(privatesize);
        kserver.public := BigNum.ModularPower (generator, kserver.private, prime);
        WriteString ("Server private key is");  WriteLn;
        BigNum.WriteBignum (kserver.private);  WriteLn;
        WriteString ("Server public key is");  WriteLn;
        BigNum.WriteBignum (kserver.public);  WriteLn;

        (* Server calculates shared key. *)

        Kshareds := BigNum.ModularPower (kclient.public, kserver.private, prime);
        WriteString ("Shared key computed by server");  WriteLn;
        BigNum.WriteBignum (Kshareds);  WriteLn;

        (* Client calculates shared key. *)

        Ksharedc := BigNum.ModularPower (kserver.public, kclient.private, prime);
        WriteString ("Shared key computed by client");  WriteLn;
        BigNum.WriteBignum (Ksharedc);  WriteLn;

        (* Next, the server generates something referred to as the      *)
        (* exchange hash H and signs it generating HS.                  *)

    END Simulate;

(************************************************************************)

BEGIN
    Oakley;        (* define some constants. *)
    Simulate (2);
    Simulate (14);
FINALLY
    BigNum.Discard (Oakley14);
    BigNum.Discard (Oakley2);
    BigNum.Discard (generator);
END testDH.

