(**************************************************************************)
(*                                                                        *)
(*  PMOS library                                                          *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE VarStrings;

        (********************************************************)
        (*                                                      *)
        (*                Variable-length strings               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        1 April 2021                    *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT ADDRESS, ADR, CARD8;

FROM BigNum IMPORT
    (* type *)  BN,
    (* proc *)  BNtoBytes;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteLn;

(************************************************************************)

CONST
    Nul = CHR(0);  LF = CHR(10);  CR = CHR(13);

(************************************************************************)

PROCEDURE WriteVarString (VAR (*IN*) str: ARRAY OF CHAR;  N: CARDINAL);

    (* Write character string of length N to standard output.   *)
    (* Convention: we treat LF as the newline character, and    *)
    (* ignore any CR encountered.                               *)

    VAR j: CARDINAL;
        ch: CHAR;

    BEGIN
        j := 0;
        WHILE j < N DO
            ch := str[j];  INC(j);
            IF ch = Nul THEN
                (* Premature termination *)
                j := N;
            ELSIF ch = CR THEN
                (* do nothing *);
            ELSIF ch = LF THEN
                WriteLn;
            ELSE
                WriteChar (ch);
            END (*IF*);
        END (*WHILE*);
    END WriteVarString;

(************************************************************************)

PROCEDURE GenStrToByteStr (p: ADDRESS;  N: CARDINAL): ByteStr;

    (* Creates a ByteStr from N bytes at address p.   *)

    VAR result: ByteStr;

    BEGIN
        WITH result DO
            start := 0;
            size := N;
            allocated := N;
            ALLOCATE (data, N);
            Copy (p, data, N);
        END (*WITH*);
        RETURN result;
    END GenStrToByteStr;

(********************************************************************************)

PROCEDURE CharStrToByteStr (str: ARRAY OF CHAR): ByteStr;

    (* Creates a ByteStr from a character string.   *)

    BEGIN
        RETURN GenStrToByteStr (ADR(str), LENGTH(str));
    END CharStrToByteStr;

(********************************************************************************)

PROCEDURE ByteStrToStr (VAR (*INOUT*) src: ByteStr;  VAR (*OUT*) result: ARRAY OF CHAR);

    (* Converts ByteStr to ordinary character string, deallocates the source.   *)

    CONST Nul = CHR(0);

    BEGIN
        IF src.size > 0 THEN
            Copy (src.data, ADR(result), src.size);
        END (*IF*);
        result[src.size] := Nul;
        IF src.allocated > 0 THEN
            DEALLOCATE (src.data, src.allocated);
        END (*IF*);
    END ByteStrToStr;

(********************************************************************************)

PROCEDURE BNtoByteStr(A: BN): ByteStr;

    (* Converts a Bignum to a ByteStr.  Does not deallocate the source. *)

    VAR N, offset: CARDINAL;
        result: ByteStr;
        bytearray: ARRAY [0..4095] OF CARD8;

    BEGIN
        N := BNtoBytes (A, bytearray);
        result.start := 0;
        result.size := N;
        offset := 0;
        IF bytearray[0] >= 80H THEN
            offset := 1;
            INC (result.size);
        END (*IF*);
        result.allocated := result.size;
        ALLOCATE (result.data, result.allocated);
        IF offset > 0 THEN
            result.data^[0] := 0;
        END (*IF*);
        Copy (ADR(bytearray), ADR(result.data^[offset]), N);
        RETURN result;
    END BNtoByteStr;

(********************************************************************************)

END VarStrings.

