MODULE MakeS;

        (****************************************************************)
        (*                                                              *)
        (*   This is a program to convert the S-boxes in the DES        *)
        (*   and convert them into more convenient lookup tables.       *)
        (*                                                              *)
        (*                                                              *)
        (*  Programmer:         P. Moylan                               *)
        (*  Started:            28 April 2018                           *)
        (*  Last edited:        28 April 2018                           *)
        (*  Status:             Working                                 *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT CARD8, CARD16, CARD32;

FROM LowLevel IMPORT
    (* proc *)  IANDB, LS, RS;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  DeleteFile, OpenNewFile, CloseFile,
                FWriteChar, FWriteString, FWriteLn, FWriteHexByte;

(************************************************************************)
(*                       WRITE HEX TO OUTPUT FILE                       *)
(************************************************************************)

PROCEDURE WriteHalfWord (cid: ChanId;  w: CARD16);

    (* Writes w in hexadecimal. *)

    BEGIN
        FWriteHexByte (cid, w DIV 256);
        FWriteHexByte (cid, w MOD 256);
    END WriteHalfWord;

(************************************************************************)

PROCEDURE WriteHexWord (cid: ChanId;  w: CARD32);

    (* Writes w in hexadecimal. *)

    BEGIN
        WriteHalfWord (cid, w DIV 65536);
        WriteHalfWord (cid, w MOD 65536);
    END WriteHexWord;

(************************************************************************)
(*                               SBOXES                                 *)
(*  The declarations here are obsolete, but I am planning to replace    *)
(*  them with a different kind of SBox.                                 *)
(************************************************************************)

TYPE SBox = ARRAY [0..63] OF CARDINAL;

(************************************************************************)
(*                           THE S PERMUTATIONS                         *)
(* Each of the 4x16 arrays below is used to map a 6-bit number to a     *)
(* 4-bit result, as shown in function Spermute below.                   *)
(************************************************************************)

TYPE
    Smap = ARRAY [0..3] OF ARRAY [0..15] OF CARD8;
    SArray = ARRAY [0..7] OF Smap;

CONST
    Sfunc = SArray{
                    (* S1 in standard *)
                    {
                       {14, 4, 13, 1, 2, 15, 11, 8, 3, 10, 6, 12, 5, 9, 0, 7},
                       {0, 15, 7, 4, 14, 2, 13, 1, 10, 6, 12, 11, 9, 5, 3, 8},
                       {4, 1, 14, 8, 13, 6, 2, 11, 15, 12, 9, 7, 3, 10, 5, 0},
                       {15, 12, 8, 2, 4, 9, 1, 7, 5, 11, 3, 14, 10, 0, 6, 13}
                    },

                    (* S2 in standard *)
                    {
                       {15, 1, 8, 14, 6, 11, 3, 4, 9, 7, 2, 13, 12, 0, 5, 10},
                       {3, 13, 4, 7, 15, 2, 8, 14, 12, 0, 1, 10, 6, 9, 11, 5},
                       {0, 14, 7, 11, 10, 4, 13, 1, 5, 8, 12, 6, 9, 3, 2, 15},
                       {13, 8, 10, 1, 3, 15, 4, 2, 11, 6, 7, 12, 0, 5, 14, 9}
                    },

                    (* S3 in standard *)
                    {
                       {10, 0, 9, 14, 6, 3, 15, 5, 1, 13, 12, 7, 11, 4, 2, 8},
                       {13, 7, 0, 9, 3, 4, 6, 10, 2, 8, 5, 14, 12, 11, 15, 1},
                       {13, 6, 4, 9, 8, 15, 3, 0, 11, 1, 2, 12, 5, 10, 14, 7},
                       {1, 10, 13, 0, 6, 9, 8, 7, 4, 15, 14, 3, 11, 5, 2, 12}
                    },

                    (* S4 in standard *)
                    {
                       {7, 13, 14, 3, 0, 6, 9, 10, 1, 2, 8, 5, 11, 12, 4, 15},
                       {13, 8, 11, 5, 6, 15, 0, 3, 4, 7, 2, 12, 1, 10, 14, 9},
                       {10, 6, 9, 0, 12, 11, 7, 13, 15, 1, 3, 14, 5, 2, 8, 4},
                       {3, 15, 0, 6, 10, 1, 13, 8, 9, 4, 5, 11, 12, 7, 2, 14}
                    },

                    (* S5 in standard *)
                    {
                       {2, 12, 4, 1, 7, 10, 11, 6, 8, 5, 3, 15, 13, 0, 14, 9},
                       {14, 11, 2, 12, 4, 7, 13, 1, 5, 0, 15, 10, 3, 9, 8, 6},
                       {4, 2, 1, 11, 10, 13, 7, 8, 15, 9, 12, 5, 6, 3, 0, 14},
                       {11, 8, 12, 7, 1, 14, 2, 13, 6, 15, 0, 9, 10, 4, 5, 3}
                    },

                    (* S6 in standard *)
                    {
                       {12, 1, 10, 15, 9, 2, 6, 8, 0, 13, 3, 4, 14, 7, 5, 11},
                       {10, 15, 4, 2, 7, 12, 9, 5, 6, 1, 13, 14, 0, 11, 3, 8},
                       {9, 14, 15, 5, 2, 8, 12, 3, 7, 0, 4, 10, 1, 13, 11, 6},
                       {4, 3, 2, 12, 9, 5, 15, 10, 11, 14, 1, 7, 6, 0, 8, 13}
                    },

                    (* S7 in standard *)
                    {
                       {4, 11, 2, 14, 15, 0, 8, 13, 3, 12, 9, 7, 5, 10, 6, 1},
                       {13, 0, 11, 7, 4, 9, 1, 10, 14, 3, 5, 12, 2, 15, 8, 6},
                       {1, 4, 11, 13, 12, 3, 7, 14, 10, 15, 6, 8, 0, 5, 9, 2},
                       {6, 11, 13, 8, 1, 4, 10, 7, 9, 5, 0, 15, 14, 2, 3, 12}
                    },

                    (* S8 in standard *)
                    {
                       {13, 2, 8, 4, 6, 15, 11, 1, 10, 9, 3, 14, 5, 0, 12, 7},
                       {1, 15, 13, 8, 10, 3, 7, 4, 12, 5, 6, 11, 0, 14, 9, 2},
                       {7, 11, 4, 1, 9, 12, 14, 2, 0, 6, 10, 13, 15, 3, 5, 8},
                       {2, 1, 14, 7, 4, 10, 8, 13, 15, 12, 9, 0, 3, 5, 6, 11}
                    }
                  };

(************************************************************************)

PROCEDURE Spermute (value: CARD8;  map: Smap): CARD8;

    (* Maps a 6-bit number to a 4-bit result.  See page 19 of standard. *)

    VAR i, j: CARD8;

    BEGIN
        i := IANDB(value, 1);
        IF IANDB (value, 20H) <> 0 THEN INC(i, 2) END(*IF*);
        j := IANDB (RS (value, 1), 0FH);
        RETURN map[i,j];
    END Spermute;

(************************************************************************)
(*                            MAKING THE TABLE                          *)
(************************************************************************)

PROCEDURE MakeTable (cid: ChanId;  N: CARDINAL);

    (* Creates the Nth box. *)

    CONST leader = "                  ";

    VAR i, k0, result: CARDINAL;

    BEGIN
        FWriteString (cid, leader);  FWriteChar (cid, '{');  FWriteLn(cid);
        FOR k0 := 0 TO 63 BY 8 DO
            FWriteString (cid, leader);
            FOR i := 0 TO 7 DO
                result := LS (Spermute (k0+i, Sfunc[N]), 28-4*N);
                IF N = 0 THEN FWriteChar (cid, '0') END(*IF*);
                WriteHexWord (cid, result);  FWriteString (cid, 'H');
                IF i < 7 THEN
                    FWriteString (cid, ', ')
                END (*IF*);
            END (*FOR*);
            IF k0 < 56 THEN
                FWriteChar (cid, ',')
            ELSE
                FWriteChar (cid, '}');
            END (*IF*);
            FWriteLn(cid);
        END (*FOR*);
        IF N < 7 THEN
            FWriteString (cid, leader);  FWriteChar (cid, ',')
        END(*IF*);
        FWriteLn(cid);
    END MakeTable;

(************************************************************************)

PROCEDURE DoTheConversion;

    CONST fname = "outdata.txt";

    VAR j: CARDINAL;
        cid: ChanId;

    BEGIN
        DeleteFile (fname);
        cid := OpenNewFile (fname, FALSE);
        FWriteString (cid, "CONST");  FWriteLn(cid);
        FWriteString (cid, "    SVal = STable{");  FWriteLn (cid);
        FOR j := 0 TO 7 DO
            MakeTable (cid, j);
        END (*FOR*);
        FWriteString (cid, "                 };");
        FWriteLn(cid);
        CloseFile (cid);
    END DoTheConversion;

(************************************************************************)

BEGIN
    DoTheConversion;
END MakeS.

